# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import inspect
import logging
import os
import sys
import time
from enum import Enum
from typing import Dict, Optional, List, Hashable, Any, Callable

import yaml

from requre.constants import (
    METATADA_KEY,
    ENV_REQURE_STORAGE_MODE,
    ENV_STORAGE_FILE,
    VERSION_REQURE_FILE,
    KEY_MINIMAL_MATCH,
)
from requre.exceptions import (
    PersistentStorageException,
    ItemNotInStorage,
    StorageNoResponseLeft,
)
from requre.utils import StorageMode

# use this sleep to avoid decorating original time function used internally
original_sleep = time.sleep
original_time = time.time
logger = logging.getLogger(__name__)


class CassetteExecution:
    """
    Execution wrapper for cassette decorators
    """

    def __init__(self):
        self._function = None
        self._cassette = None
        self._obj_cls = None

    @property
    def function(self):
        """
        return stored function
        """
        return self._function

    @function.setter
    def function(self, value):
        self._function = value

    @property
    def cassette(self):
        """
        return current cassette for function execution
        """
        return self._cassette

    @cassette.setter
    def cassette(self, value):
        self._cassette = value

    def __call__(self, *args, **kwargs):
        """
        backward compatibility and for usage within inport system.
        please use this carrefully, in case of object instance it removes self
        """
        # TODO: fix issue with self removing
        return self.function(*args, **kwargs)

    @property
    def obj_cls(self):
        return self._obj_cls

    @obj_cls.setter
    def obj_cls(self, value):
        self._obj_cls = value


def _get_module_name(obj: Any) -> Optional[str]:
    module = inspect.getmodule(obj)
    if module:
        return module.__name__
    elif (
        hasattr(obj, "__class__")
        and obj.__class__.__name__ == "builtin_function_or_method"
    ):
        return obj.__class__.__name__
    else:
        return ""


class StorageKeysInspect:
    @staticmethod
    def get_base_keys(func: Callable) -> List[Any]:
        raise NotImplementedError("Use child classes")


class StorageKeysInspectFull(StorageKeysInspect):
    @staticmethod
    def get_base_keys(func: Callable) -> List[Any]:
        output: List[str] = list()
        # callers module list, to be able to separate requests for various services in one file
        caller_list: List[str] = list()
        for currnetframe in inspect.stack():
            module_name = _get_module_name(currnetframe[0])
            if not module_name:
                continue
            if module_name.startswith("_"):
                break
            # avoid to store requre.storage to module stack
            # backward compatibility issue
            if module_name.startswith("requre.storage"):
                continue
            if len(caller_list) and caller_list[-1] == module_name:
                continue
            else:
                caller_list.append(module_name)
        output += caller_list[::-1]
        # module name where function is
        output.append(_get_module_name(func))
        # name of function what were used
        output.append(func.__name__)
        return output


class StorageKeysInspectOuter(StorageKeysInspect):
    @staticmethod
    def get_base_keys(func: Callable) -> List[Any]:
        """
        return shorter stack information what will end after info goes to CWD.
        Be careful: You have to use same scheduler (when you regenerate tests via pytest and
        then run them via tox, it causes that tox what uses python venv installs all deps to
        .tox subdir, so that .tox files will be in CWD then)
        example: full stack info for packit project:

        "long python execute handling":
          unittest.case:
            tests_recording.test_status:
              packit.status:
                ogr.services.pagure.project:
                  ogr.services.pagure.service:
                    requests.sessions:
                      requre.objects:
                        requests.sessions:
                            send:
        the current one remove all stack info before it goes to current directory (including)
                ogr.services.pagure.project:
                  ogr.services.pagure.service:
                    requests.sessions:
                      requre.objects:
                        requests.sessions:
                            send:

        """
        output: List[str] = list()
        # callers module list, to be able to separate requests for various services in one file
        caller_list: List[str] = list()
        for currnetframe in inspect.stack():
            module_name = _get_module_name(currnetframe[0])
            if not module_name:
                continue
            module_file = inspect.getmodule(currnetframe[0]).__file__
            # If python stack is already in directory you are (CWD) then stop appending
            # Because you dont want to track changes of test call stack or your project stack
            # This is main feature regarding to StorageKeysInspectFull, what stores it as well
            # and may cause issue with unittest execution changes
            if os.path.realpath(os.getcwd()) in os.path.realpath(module_file):
                break
            # avoid to store requre.storage to module stack
            # backward compatibility issue
            if module_name.startswith("requre.storage"):
                continue
            # avoid duplication in stack information, e.g. one method will can another one from
            # same file/module, so avoid to add it to list, in case already here on last position.
            if len(caller_list) and caller_list[-1] == module_name:
                continue
            else:
                caller_list.append(module_name)
        output += caller_list[::-1]
        # module name where function is
        output.append(_get_module_name(func))
        # name of function what were used
        output.append(func.__name__)
        return output


class StorageKeysInspectSimple(StorageKeysInspect):
    @staticmethod
    def get_base_keys(func: Callable) -> List[Any]:
        return [_get_module_name(func), func.__name__]


class StorageKeysInspectUnique(StorageKeysInspect):
    @staticmethod
    def get_base_keys(func: Callable) -> List[Any]:
        return StorageKeysInspectUnique._get_unique_keys(
            StorageKeysInspectFull.get_base_keys(func)
        )

    @staticmethod
    def _get_unique_keys(keys):
        output: List = []
        for item in reversed(keys):
            if item not in output:
                output.append(item)
        return list(reversed(output))


StorageKeysInspectDefault = StorageKeysInspectFull


class DataStructure:
    """
    Object model for storing data to persistent storage
    """

    OUTPUT_KEY = "output"
    METADATA_KEY = "metadata"

    def __init__(self, output: Any):
        self.output = output
        self._metadata: dict = dict()

    @property
    def metadata(self):
        """
        Medatata for item in storage
        """
        return self._metadata

    @metadata.setter
    def metadata(self, values: dict):
        for k, v in values.items():
            self._metadata[k] = v

    def dump(self):
        """
        Dump object to serializable format

        :return dict
        """
        return {self.METADATA_KEY: self.metadata, self.OUTPUT_KEY: self.output}

    @classmethod
    def create_from_value(cls, dict_repr: dict):
        """
        Create Object representation from dict

        :return DataStructure
        """
        data = cls(dict_repr[cls.OUTPUT_KEY])
        data.metadata = dict_repr[cls.METADATA_KEY]
        return data

    @classmethod
    def create_from_dict(cls, dict_repr: dict, cassette: Any):
        """
        Create Object representation from dict

        :param cassette: Cassette instance to pass inside object to work with
        :return DataStructure
        """
        if cassette.data_miner.key not in dict_repr:
            raise ItemNotInStorage(
                f"Key '{cassette.data_miner.key}' not in the response file. "
                f"(Be sure that you generate the response file "
                f"for this variant.)"
            )
        value = dict_repr[cassette.data_miner.key]
        data = cls(value[cls.OUTPUT_KEY])
        data.metadata = value[cls.METADATA_KEY]
        return data

    @classmethod
    def create_from_dict_with_list(cls, dict_repr: dict, cassette: Any):
        """
        Create Object representation from dict with list

        :param cassette: Cassette instance to pass inside object to work with
        :return DataStructure
        """
        if cassette.data_miner.key not in dict_repr:
            raise ItemNotInStorage(
                f"Key '{cassette.data_miner.key}' not in the response file. "
                f"(Be sure that you generate the response file "
                f"for this variant.)"
            )
        value = dict_repr[cassette.data_miner.key]
        return cls.create_from_list(list_repr=value, cassette=cassette)

    @staticmethod
    def create_from_list(list_repr: list, cassette: Any):
        if len(list_repr) == 0:
            raise StorageNoResponseLeft(
                f"Unable to find response (file: {cassette.storage_file}). "
                "You are in read-only mode now, that's why you are seeing this error message. "
                "You should record the test case again "
                "by removing the response file and running the test(s) again."
            )
        if cassette.storage_file_version <= 1:
            # Backward compatibility for requre before using DataStructure
            data = DataStructure(list_repr[0])
        else:
            data = DataStructure.create_from_value(list_repr[0])
        del list_repr[0]
        return data


class DataTypes(Enum):
    List = 1
    Value = 2
    Dict = 3
    DictWithList = 4


class DataMiner:
    """
    Intermediate class used for proper formatting and keeping backward
    compatibility with requre storage version files
    """

    def __init__(self):
        self.current_time = original_time()
        self.data: Optional[DataStructure] = None
        self.data_type: DataTypes = DataTypes.List
        self.use_latency = False
        self.LATENCY_KEY = "latency"
        self.key: str = "all"
        self.key_stategy_cls = StorageKeysInspectDefault
        self.store_arg_debug_metadata = False
        self.METADATA_ARG_DEBUG_KEY = "log_call_function"
        self.METADATA_CALLER_LIST = "module_call_list"
        self.read_key_exact = False

    def get_latency(self, regenerate=True) -> float:
        """
        Returns latency from last store event

        :param regenerate: if False, then it will not be changed, just displayed
        :return: float
        """
        current_time = original_time()
        diff = current_time - self.current_time
        if regenerate:
            self.current_time = current_time
        return diff

    def dump(self, level, key, values, metadata: Dict, cassette: Any):
        """
        Store values in proper format to level[key] item

        :param level: parent dict object
        :param key: item in dict
        :param values: what to store (return value of function)
        :param metadata: store metadata to object from upper level
        :param cassette: Cassette instance to pass inside object to work with
        :return: None
        """
        ds = DataStructure(values)
        if metadata:
            ds.metadata = metadata
        if self.LATENCY_KEY not in ds.metadata:
            ds.metadata = {self.LATENCY_KEY: self.get_latency()}
        self.data = ds
        item = ds.dump()
        if self.data_type == DataTypes.Dict:
            level.setdefault(key, {})
            level[key][self.key] = item
        elif self.data_type == DataTypes.DictWithList:
            level.setdefault(key, {})
            level[key].setdefault(self.key, [])
            level[key][self.key].append(item)
        elif self.data_type in [DataTypes.List, DataTypes.DictWithList]:
            if cassette.storage_file_version <= 1:
                # Backward compatibility for requre before using DataStructure
                item = values
            level.setdefault(key, [])
            level[key].append(item)
        elif self.data_type == DataTypes.Dict:
            level.setdefault(key, {})
            level[key][self.key] = item
        else:
            level[key] = item

    def load(self, level, cassette: Any):
        """
        Get data from storage_object and trasform it to DataStructure object.

        :param level: where data are stored
        :param cassette: Cassette instance to pass inside object to work with
        :return: output of function
        """
        data = None
        if self.data_type == DataTypes.List:
            data = DataStructure.create_from_list(level, cassette=cassette)
        elif self.data_type == DataTypes.Value:
            data = DataStructure.create_from_value(level)
        elif self.data_type == DataTypes.Dict:
            data = DataStructure.create_from_dict(level, cassette=cassette)
        elif self.data_type == DataTypes.DictWithList:
            data = DataStructure.create_from_dict_with_list(level, cassette=cassette)
        self.data = data
        if self.use_latency:
            original_sleep(data.metadata.get(self.LATENCY_KEY, 0))
        return data.output

    @property
    def metadata(self):
        return self.data.metadata

    @metadata.setter
    def metadata(self, value):
        self.data.metadata = value


class Cassette:
    """
    Class implements reading/writing simple JSON requests to dict structure
    and return values based on keys.
    It contains methods to reads/stores data to object and load and store them to YAML file

    storage_object: dict with structured data based on keys (eg. simple JSON requests)
    storage_file: file for reading and writing data in storage_object
    """

    @property
    def content(self) -> dict:
        return self.storage_object

    internal_object_key = METATADA_KEY
    version_key = "version_storage_file"
    key_inspect_strategy_key = "key_strategy"

    def _set_defaults(self) -> None:
        self.dump_after_store = False
        self.is_flushed = False
        self.storage_object: dict = {}
        self._storage_file: Optional[str] = None
        self.data_miner = DataMiner()
        self.mode = StorageMode.default

    def __init__(self) -> None:
        # call dump() after store() is called
        self._set_defaults()
        storage_file_from_env = os.getenv(ENV_STORAGE_FILE)
        if storage_file_from_env:
            self.storage_file = storage_file_from_env

    def __str__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"storage_mode={self.mode}, "
            f"storage_file={self._storage_file})"
        )

    def __repr__(self) -> str:
        return self.__str__()

    @property
    def metadata(self):
        """
        Placeholder to store some custom configuration, or store custom data,
        eg. version of storage file, or some versions of packages

        :return: dict
        """
        return self.storage_object.get(self.internal_object_key, {})

    @metadata.setter
    def metadata(self, key_dict: dict):
        if self.internal_object_key not in self.storage_object:
            self.storage_object[self.internal_object_key] = {}
        for k, v in key_dict.items():
            self.storage_object[self.internal_object_key][k] = v
        self.is_flushed = False

    @property
    def storage_file_version(self):
        """
        Get version of persistent storage file
        :return: int
        """
        return self.metadata.get(self.version_key, 0)

    def do_store(self, keys):
        if self.mode == StorageMode.read and not os.path.exists(self.storage_file):
            raise PersistentStorageException(
                "Requre can't work in this setup: we are meant to read "
                "recorded responses but the storage file does not exist."
            )
        if self.mode == StorageMode.read:
            return False

        elif self.mode == StorageMode.write:
            return True

        elif self.mode == StorageMode.append:
            if self.data_miner.data_type in [DataTypes.DictWithList, DataTypes.List]:
                # append possible
                return True

            if (
                self.data_miner.data_type in [DataTypes.Dict, DataTypes.Value]
                and keys not in self
            ):
                # keys present => we can read the values
                return True
            return False

        key_present = keys in self
        logger.warning(
            f"No storage-mode set. Checking presence of the keys: {key_present} "
            f"=> {'read' if key_present else 'store'}"
        )
        return not key_present

    @storage_file_version.setter  # type: ignore
    def storage_file_version(self, value: int):
        self.metadata = {self.version_key: value}

    def _set_storage_metadata_if_not_set(self):
        """
        Set storage file version if not already set to latest version.
        """
        if self.metadata.get(self.version_key) is None:
            self.storage_file_version = VERSION_REQURE_FILE
        if self.metadata.get(self.key_inspect_strategy_key) is None:
            self.metadata = {
                self.key_inspect_strategy_key: self.data_miner.key_stategy_cls.__name__
            }
        data_type_name = DataTypes.__name__
        if self.metadata.get(data_type_name) is None:
            self.metadata = {data_type_name: self.data_miner.data_type.value}

    def _set_storage_mode(self, storage_file):
        # use env var mode if given as the most important
        storage_mode = os.getenv(ENV_REQURE_STORAGE_MODE)
        if storage_mode:
            logger.info(
                f"You overrided storage mode via env var {ENV_REQURE_STORAGE_MODE}={storage_mode}"
            )
            if not hasattr(StorageMode, storage_mode):
                raise PersistentStorageException(
                    f"storage mode '{storage_mode}' does not exist, "
                    f"use one of {list(StorageMode.__members__.keys())}) "
                )
            self.mode: StorageMode = getattr(StorageMode, storage_mode)

        # if not given by ENV VAR, guess mode via file existence
        else:
            if not os.path.exists(storage_file):
                if self.mode == StorageMode.default:
                    self.mode = StorageMode.write
            else:
                if self.mode == StorageMode.default:
                    self.mode = StorageMode.read
                self.storage_object = self.load()
        if self.mode == StorageMode.read and not os.path.exists(storage_file):
            raise PersistentStorageException(
                "Requre can't work in this setup: we are meant to read"
                f" recorded responses but the storage file ({storage_file}) "
                "does not exist."
            )

    @property
    def storage_file(self):
        return self._storage_file

    @storage_file.setter
    def storage_file(self, value):
        # when set to None, reset to default
        logger.info(f"SETTING SF: {value} for cassette {id(self)} - {self}")
        if value is None:
            self._set_defaults()
            return
        # when file is changed set PersistenStorage default values
        if self._storage_file != value:
            self._set_defaults()
            self._storage_file = value
        self._set_storage_mode(storage_file=self._storage_file)
        # load data for read.
        if self.mode == StorageMode.read:
            self.storage_object = self.load()
        # if append mode load data just in case file exists.
        if self.mode == StorageMode.append and os.path.exists(self._storage_file):
            self.storage_object = self.load()

    @staticmethod
    def transform_hashable(keys: List) -> List:
        output: List = []
        for item in keys:
            if not item:
                output.append("empty")
            elif not isinstance(item, Hashable):
                output.append(str(item))
            else:
                output.append(item)
        return output

    def _pretty_dict_output(self, input_dict: Optional[dict] = None, depth=0, sep="  "):
        if not isinstance(input_dict, dict):
            return
        for key, value in input_dict.items():
            yield (sep * depth) + str(key)
            yield from self._pretty_dict_output(value, depth + 1)

    def _printable_dict_output(self, input_dict: dict):
        return (
            "Current internal object structure:"
            + "\n"
            + "\n".join(list(self._pretty_dict_output(self.storage_object)))
        )

    def store(self, keys: List, values: Any, metadata: Dict) -> None:
        """
        Stores data to dictionary object based on keys values it will create structure
        if structure does not exist

        It implicitly changes type to string if key is not hashable

        :param keys: items what will be used as keys for dictionary
        :param values: It could be whatever type what is used in original object handling
        :return: None
        """
        self._set_storage_metadata_if_not_set()
        current_level = self.storage_object
        level_trace = []
        hashable_keys = self.transform_hashable(keys)
        for item_num in range(len(hashable_keys)):
            item = hashable_keys[item_num]
            level_trace.append(item)
            if item_num + 1 < len(hashable_keys):
                if not isinstance(current_level, dict):
                    print(
                        self._printable_dict_output(self.storage_object),
                        file=sys.stderr,
                    )
                    raise PersistentStorageException(
                        "you are mixing various depths of stored data: keys:"
                        f" {hashable_keys}, current levels: {level_trace}"
                    )
                if not current_level.get(item):
                    current_level[item] = {}
            else:
                self.data_miner.dump(
                    level=current_level,
                    key=item,
                    values=values,
                    metadata=metadata,
                    cassette=self,
                )
            current_level = current_level[item]
        self.is_flushed = False

        if self.dump_after_store:
            self.dump()
        logger.debug(f"Storing response to: {self.storage_file}: {hashable_keys}")

    def read(self, keys: List) -> Any:
        """
        Reads data from dictionary object structure based on keys.
        If keys does not exists

        It implicitly changes type to string if key is not hashable

        :param keys: key list for searching in dict
        :return: value assigged to key items
        """
        current_level = self.storage_object
        hashable_keys = self.transform_hashable(keys)
        debug_keys: List[str] = []
        matched_calls: List[str] = []
        list_len = len(hashable_keys)
        for item_num in range(list_len):
            item = hashable_keys[item_num]
            if item not in current_level:
                # it matched last 2 items
                if self.data_miner.read_key_exact or (
                    not self.data_miner.read_key_exact
                    and item_num + KEY_MINIMAL_MATCH >= list_len
                ):
                    # if not matched, but consider if it is not same key as previous
                    # it is important if simplify used.
                    if matched_calls and item == matched_calls[-1]:
                        debug_keys.append(f"DUPLICATE {item}")
                        continue
                    print(
                        self._printable_dict_output(self.storage_object),
                        file=sys.stderr,
                    )
                    raise ItemNotInStorage(
                        f"Keys not in storage:{self.storage_file}"
                        f" Matched: {debug_keys},"
                        f" Missing: {hashable_keys[item_num:]}"
                    )
                else:
                    debug_keys.append(f"SKIP {item}")

            else:
                debug_keys.append(item)
                matched_calls.append(item)
                current_level = current_level[item]
        try:
            result = self.data_miner.load(level=current_level, cassette=self)
        except StorageNoResponseLeft as e:
            raise StorageNoResponseLeft(f"{e.args} (keys: {debug_keys})")
        logger.debug(f"Reading response from: {self.storage_file}: {debug_keys}")
        return result

    def __contains__(self, item) -> bool:
        current_level = self.storage_object
        hashable_keys = self.transform_hashable(item)
        for item in hashable_keys:
            if item not in current_level:
                return False
            current_level = current_level[item]
        if self.data_miner.data_type in [DataTypes.Dict, DataTypes.DictWithList]:
            return self.data_miner.key in current_level
        return True

    def __getitem__(self, key):
        return self.read(keys=key)

    def __setitem__(self, key, value):
        self.store(keys=key, values=value, metadata={})

    def __delitem__(self, key):
        current_level = self.storage_object
        hashable_keys = self.transform_hashable(key)
        last_level = None
        for item in hashable_keys:
            if item not in current_level:
                raise ItemNotInStorage(
                    f"Key not found in the persistent storage: {key}"
                )
            last_level = current_level
            current_level = current_level[item]

        if self.data_miner.data_type == DataTypes.Dict:
            del current_level[self.data_miner.key]
        else:
            del last_level[key[-1]]

    def dump(self) -> None:
        """
        Explicitly stores content of storage_object to storage_file path

        This method is also called when object is deleted and is set write mode to True

        :return: None
        """
        if self.mode in [StorageMode.write, StorageMode.append]:
            self._set_storage_metadata_if_not_set()
            if self.is_flushed:
                return None
            with open(self.storage_file, "w") as yaml_file:
                yaml.dump(self.storage_object, yaml_file, default_flow_style=False)
            self.is_flushed = True

    def load(self) -> Dict:
        """
        Explicitly loads file content of storage_file to storage_object and return as well

        :return: dict
        """
        with open(self.storage_file, "r") as yaml_file:
            output = yaml.safe_load(yaml_file)
        self.storage_object = output
        # set proper storage strategy if stored in file
        if self.metadata.get(self.key_inspect_strategy_key):
            logger.debug(
                f"Used key strategy: {self.metadata.get(self.key_inspect_strategy_key)}"
            )
            self.data_miner.key_stategy_cls = globals()[
                self.metadata.get(self.key_inspect_strategy_key)
            ]
        return output



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from unittest import TestCase
from requre.cassette import CassetteExecution


class Execution(TestCase):
    def testCreate(self):
        ce = CassetteExecution()
        ce.function = lambda: "ahoj"
        ce.cassette = "nothing"
        self.assertEqual("ahoj", ce.function())
        self.assertEqual("ahoj", ce())



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from tests.testbase import BaseClass
import tempfile
from requre.simple_object import Simple
from requre.online_replacing import replace


def x():
    return tempfile.mktemp()


class Duplicated(BaseClass):
    @replace("tests.test_duplication.x", decorate=Simple.decorator_plain())
    @replace("tempfile.mktemp", decorate=Simple.decorator_plain())
    def test(self):
        a = x()
        b = tempfile.mktemp()
        self.assertNotEqual(a, b)



from requre.objects import ObjectStorage
from requre.cassette import Cassette

cassette = Cassette()


@ObjectStorage.decorator_plain(cassette=cassette)
def return_result():
    return {"value": input("insert value: ")}


cassette.storage_file = "values.yaml"
value = return_result()

cassette.dump()
print(cassette.storage_object)
print(value)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from requre.helpers import record_requests as _record_requests
from requre.helpers import recording_requests as _recording_requests
from requre.record_and_replace import (
    apply_decorator_to_all_methods as _apply_decorator_to_all_methods,
)
from requre.record_and_replace import record as _record
from requre.record_and_replace import recording as _recording
from requre.record_and_replace import replace

# BACKWARDS COMPATIBILITY
recording = _recording
record = _record
replace_module_match = replace
record_requests = _record_requests
record_requests_for_all_methods = record_requests
recording_requests = _recording_requests
apply_decorator_to_all_methods = _apply_decorator_to_all_methods



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import os
from typing import Any, Callable

from git.refs.head import HEAD
from git.remote import PushInfo
from git.repo.base import Repo
from git.util import IterableList

from requre.objects import ObjectStorage
from requre.helpers.files import StoreFiles


class PushInfoStorageList(ObjectStorage):
    __ignored = ["_remote"]
    __response_keys_special = ["local_ref"]
    __response_keys = list(
        set(PushInfo.__slots__) - set(__ignored) - set(__response_keys_special)
    )
    stack_internal_check = False

    object_type = IterableList

    def to_serializable(self, obj: Any) -> Any:
        output = list()
        for item in obj:
            tmp = dict()
            for key in self.__response_keys:
                tmp[key] = getattr(item, key)
            for key in self.__response_keys_special:
                if key == "local_ref":
                    tmp[key] = [
                        getattr(item, key).repo.git_dir,
                        getattr(item, key).path,
                    ]
            output.append(tmp)
        return output

    def from_serializable(self, data: Any) -> Any:
        out = self.object_type("name")
        for item in data:
            tmp = PushInfo(None, None, None, None)
            for key in self.__response_keys:
                setattr(tmp, key, item[key])
            for key in self.__response_keys_special:
                if key == "local_ref":
                    setattr(tmp, key, HEAD(Repo(item[key][0])))
            out.append(tmp)
        return out

    @classmethod
    def execute(cls, keys: list, func: Callable, *args, **kwargs) -> Any:
        """
        Class method to store or read object from persistent storage
        with explicit set of *args, **kwargs parameters to use as keys
        :param keys: list of keys used as parameter
        :param func: original function
        :param args: parameters of original function
        :param kwargs: parameters of original function
        :return: output of called func
        """

        output = super().execute(keys, func, *args, **kwargs)
        git_object = args[0]
        remote_url = git_object.repo.remotes[git_object.name].url
        if os.path.isdir(remote_url):
            StoreFiles.explicit_reference(remote_url)
        return output



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest

from requre.utils import get_datafile_filename


@pytest.fixture
def remove_storage_file(request):
    storage_file = get_datafile_filename(request.node)
    if storage_file.is_file():
        storage_file.unlink()
    return storage_file


@pytest.fixture
def remove_storage_file_after(request):
    storage_file = get_datafile_filename(request.node)
    yield storage_file
    # TODO: found why this were necessary before change of singletons
    # PersistentObjectStorage().cassette.dump()
    if storage_file.is_file():
        storage_file.unlink()



from requre.helpers.files import StoreFiles
from requre.cassette import Cassette
from tempfile import mkdtemp
import os

cassette = Cassette()
cassette.storage_file = "/tmp/files.yaml"
temp_dir = mkdtemp()


@StoreFiles.where_arg_references(
    key_position_params_dict={"dir_name": 0}, cassette=cassette
)
def return_result(dir_name):
    file_name = input("enter file name: ")
    with open(os.path.join(dir_name, file_name), "w") as fd:
        fd.write("empty")
    return file_name


print("returned file name:", return_result(temp_dir))
print("dir name (current):", temp_dir)
print("files:", os.listdir(temp_dir))
cassette.dump()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import math
import os
import socket
import unittest
from typing import Optional

import requests

import tests.data.special_requre_module
from requre.cassette import Cassette, StorageMode
from requre.helpers.requests_response import RequestResponseHandling
from requre.simple_object import Simple
from requre.online_replacing import (
    apply_decorator_to_all_methods,
    record,
    record_requests,
    replace_module_match,
)
from tests.data import special_requre_module
from tests.data.special_requre_module import hello


def guard(*args, **kwargs):
    raise IOError("No Internet connection")


original_socket = socket.socket

SELECTOR = str(os.path.basename(__file__).rsplit(".", 1)[0])
TEST_DIR = os.path.dirname(__file__)


def decorator_exact(func):
    def _int():
        return "decorator_c " + func()

    return _int


def decorator_inc(func):
    def _int_inc(a):
        return func(a) + 1

    return _int_inc


@replace_module_match(
    what="tests.data.special_requre_module.hello", decorate=decorator_exact
)
def decorated_exact():
    return tests.data.special_requre_module.hello() + hello() + "exact"


class TestOnlinePatchingModuleMatch(unittest.TestCase):
    def testDecoratorImport(self):
        self.assertEqual(decorated_exact(), "decorator_c Hi! decorator_c Hi! exact")
        # test reverting one decorators
        self.assertEqual(tests.data.special_requre_module.hello(), "Hi! ")
        self.assertEqual(hello(), "Hi! ")

    @replace_module_match(
        what="tests.data.special_requre_module.hello", decorate=decorator_exact
    )
    def testDecoratorMainUsage(self):
        self.assertEqual(tests.data.special_requre_module.hello(), "decorator_c Hi! ")
        self.assertEqual(hello(), "decorator_c Hi! ")
        self.assertEqual(special_requre_module.hello(), "decorator_c Hi! ")

    @replace_module_match(
        what="tests.data.special_requre_module.hello",
        decorate=[decorator_exact, decorator_exact],
    )
    @replace_module_match(
        what="tests.data.special_requre_module.inc", decorate=decorator_inc
    )
    def testDecoratorMultipleDecorators(self):
        self.assertEqual(hello(), "decorator_c decorator_c Hi! ")
        # verify the decorator
        self.assertEqual(decorator_inc(lambda x: x)(1), 2)
        # verify decorated function
        self.assertEqual(tests.data.special_requre_module.inc(1), 3)


# part of documented workaround for the next testcase
setattr(tests.data.special_requre_module.dynamic, "other", lambda self: "static")


class DynamicMethods(unittest.TestCase):
    @replace_module_match(
        what="tests.data.special_requre_module.dynamic.some", decorate=decorator_exact
    )
    def testDynamicClassMethodNotWorking(self):
        self.assertRaises(
            AttributeError, getattr, tests.data.special_requre_module.dynamic, "some"
        )
        self.assertEqual(tests.data.special_requre_module.dynamic().some(), "SOME")

    @unittest.skip("this also does not work, probably caused by pytest execution")
    @replace_module_match(
        what="tests.data.special_requre_module.dynamic.other", decorate=decorator_exact
    )
    def testDynamicClassMethodWorking(self):
        self.assertEqual(
            tests.data.special_requre_module.dynamic().other(), "decorated_c static"
        )


class CassetteSelection(unittest.TestCase):
    def setUp(self) -> None:
        # disable internet access via sockets
        setattr(socket, "socket", guard)

    def reset(self):
        setattr(socket, "socket", original_socket)

    def tearDown(self) -> None:
        self.reset()

    def testGuard(self):
        # check if
        self.assertRaises(IOError, requests.get, "http://example.com")

    @replace_module_match(
        what="requests.sessions.Session.request",
        decorate=RequestResponseHandling.decorator(item_list=["method", "url"]),
    )
    def testWrite(self, cassette: Cassette):
        self.reset()
        response = requests.get("http://example.com")
        self.assertIn("This domain is for use", response.text)
        self.assertFalse(os.path.exists(cassette.storage_file))
        cassette.dump()
        self.assertTrue(os.path.exists(cassette.storage_file))
        os.remove(cassette.storage_file)

    @replace_module_match(
        what="requests.sessions.Session.request",
        decorate=RequestResponseHandling.decorator(item_list=["method", "url"]),
    )
    def testRead(self, cassette: Cassette):
        if cassette.mode == StorageMode.read:
            self.assertTrue(os.path.exists(cassette.storage_file))
        else:
            self.reset()
        response = requests.get("http://example.com")
        self.assertIn("This domain is for use", response.text)

    @replace_module_match(
        what="requests.sessions.Session.request",
        decorate=RequestResponseHandling.decorator(item_list=["method", "url"]),
    )
    @replace_module_match(what="math.sin", decorate=Simple.decorator(item_list=[]))
    def testReadMultiple(self, cassette: Cassette):
        assert cassette
        if cassette.mode == StorageMode.write:
            self.reset()
            sin_output = math.sin(1.5)
        else:
            sin_output = math.sin(4)
        response = requests.get("http://example.com")
        self.assertIn("This domain is for use", response.text)
        self.assertAlmostEqual(0.9974949866040544, sin_output, delta=0.0005)


new_cassette = Cassette()


@apply_decorator_to_all_methods(
    replace_module_match(
        what="math.sin", decorate=Simple.decorator_plain(), cassette=new_cassette
    )
)
@record_requests(cassette=new_cassette)
class DecoratorClassApply(unittest.TestCase):
    # when regeneration, comment lines with assert equals, because checks for equality does not work
    def setUp(self):
        new_cassette.storage_file = None

    def test0(self, cassette: Cassette):
        if cassette.mode == StorageMode.read:
            self.assertEqual(len(new_cassette.storage_object["math"]["sin"]), 1)
        sin_output = math.sin(1.5)
        response = requests.get("http://example.com")
        self.assertIn("This domain is for use", response.text)
        self.assertAlmostEqual(0.9974949866040544, sin_output, delta=0.0005)

    def test1(self):
        sin_output = math.sin(1.5)
        response = requests.get("http://example.com")
        self.assertIn("This domain is for use", response.text)
        self.assertAlmostEqual(0.9974949866040544, sin_output, delta=0.0005)

    def test2(self, cassette: Cassette):
        if cassette.mode == StorageMode.read:
            self.assertEqual(len(new_cassette.storage_object["math"]["sin"]), 2)
        self.test1()
        if cassette.mode == StorageMode.read:
            self.assertEqual(len(new_cassette.storage_object["math"]["sin"]), 1)


@record(what="tests.data.special_requre_module.random_number")
class RecordDecoratorForClass(unittest.TestCase):
    def test_random(self):
        random_number = tests.data.special_requre_module.random_number()
        self.assertEqual(random_number, 0.819349292484907)


@record(what="tests.data.special_requre_module.random_number")
@record(what="tests.data.special_requre_module.another_random_number")
class RecordDecoratorForClassMultiple(unittest.TestCase):
    def test_random(self):
        random_number = tests.data.special_requre_module.random_number()
        another_random_number = tests.data.special_requre_module.another_random_number()
        self.assertEqual(0.6749983968044089, random_number)
        self.assertEqual(0.15676319106079073, another_random_number)


class RecordDecoratorForMethod(unittest.TestCase):
    @record(what="tests.data.special_requre_module.random_number")
    def test_random(self):
        random_number = tests.data.special_requre_module.random_number()
        self.assertEqual(random_number, 0.17583106733657616)


class RecordDecoratorForMethodMultiple(unittest.TestCase):
    @record(what="tests.data.special_requre_module.random_number")
    @record(what="tests.data.special_requre_module.another_random_number")
    def test_random(self):
        random_number = tests.data.special_requre_module.random_number()
        another_random_number = tests.data.special_requre_module.another_random_number()
        self.assertEqual(0.2809709312583647, random_number)
        self.assertEqual(0.039985857086601295, another_random_number)


@record(what="tests.data.special_requre_module.random_number")
class RecordDecoratorForClassCassette(unittest.TestCase):
    cassette: Optional[Cassette]

    def setUp(self) -> None:
        self.setup = False
        self.teardown = False
        self.cassette = None
        super().setUp()

    def cassette_setup(self, cassette: Cassette):
        assert cassette
        self.cassette = cassette
        self.setup = True

    def test_random(self, cassette: Cassette):
        assert cassette
        assert self.setup
        assert not self.teardown
        assert self.cassette == cassette
        random_number = tests.data.special_requre_module.random_number()
        self.assertEqual(random_number, 0.9720986758204466)

    def cassette_teardown(self, cassette: Cassette):
        assert cassette
        assert self.cassette == cassette
        self.teardown = True

    def tearDown(self) -> None:
        assert self.teardown
        super().tearDown()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import sys
import tempfile
from requre.simple_object import Tuple
from requre.online_replacing import apply_decorator_to_all_methods, replace
from requre.cassette import Cassette, StorageMode
from tests.testbase import BaseClass, RetTuple


@apply_decorator_to_all_methods(
    replace(what="tests.testbase.RetTuple.ret", decorate=Tuple.decorator_plain())
)
class TestTuple(BaseClass):
    def test_write(self, cassette: Cassette):
        cassette.storage_file = tempfile.mktemp()
        out = RetTuple().ret(1)
        self.assertEqual(out, ("ret", 1))

    def test_read(self, cassette: Cassette):
        print(cassette.storage_file)
        print([k for k, v in sys.modules.items() if "tests" in k])
        if cassette.mode == StorageMode.write:
            out = RetTuple().ret(2)
            self.assertEqual(out, ("ret", 2))
        else:
            out = RetTuple().ret(1)
            self.assertEqual(out, ("ret", 2))



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import os
import unittest

from ogr import GithubService

from requre.helpers.requests_response import (
    RequestResponseHandling,
    remove_password_from_url,
)
from requre.online_replacing import replace_module_match, record_requests
from requre.cassette import StorageMode, Cassette


class GithubTests(unittest.TestCase):
    def _pr_comments_test(self):
        token = os.environ.get("GITHUB_TOKEN")
        cassette = Cassette()
        if cassette.mode == StorageMode.write and (not token):
            raise EnvironmentError(
                f"You are in Requre write mode, please set proper GITHUB_TOKEN"
                f" env variables {cassette.storage_file}"
            )
        # possible to check before reading values because in other case values are removed
        # and in write mode is does have sense at the end
        if cassette.mode == StorageMode.read:
            self.assertIn(self.id(), cassette.storage_file)
            self.assertIn("LGTM", str(cassette.storage_object))
            self.assertTrue(
                cassette.storage_object["requests.sessions"]["send"]["GET"][
                    "https://api.github.com:443/repos/packit-service/ogr"
                ]
            )
        service = GithubService(token=token)
        ogr_project = service.get_project(namespace="packit-service", repo="ogr")
        pr_comments = ogr_project.get_pr_comments(9)
        assert pr_comments
        assert len(pr_comments) == 2

        assert pr_comments[0].body.endswith("fixed")
        assert pr_comments[1].body.startswith("LGTM")

    @replace_module_match(
        what="requests.sessions.Session.request",
        decorate=RequestResponseHandling.decorator(
            item_list=["method", "url"],
            map_function_to_item={"url": remove_password_from_url},
        ),
    )
    def test_pr_comments(self):
        self._pr_comments_test()

    @record_requests
    def test_pr_comments_record_requests_decorator(self):
        self._pr_comments_test()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import functools
import inspect
import logging
import os
import re
import sys
from contextlib import contextmanager
from typing import Any, Callable, List, Optional, Tuple, Union

from requre.cassette import Cassette, CassetteExecution
from requre.cassette import StorageKeysInspectSimple
from requre.constants import (
    REQURE_CASSETTE_ATTRIBUTE_NAME,
    REQURE_SETUP_APPLIED_ATTRIBUTE_NAME,
    TEST_METHOD_REGEXP,
)
from requre.guess_object import Guess
from requre.utils import get_datafile_filename, get_module_of_previous_context

logger = logging.getLogger(__name__)


class ModuleRecord:
    """
    Store data entry for replacement system, to be able to _revert_modules changes
    back after execution finishes.
    """

    def __init__(self, what, parent, original, replacement, add_revert_list=None):
        self.what = what
        self.parent = parent
        self.original = original
        self.replacement = replacement
        self.add_revert_list = add_revert_list or []
        self.caller_module = get_module_of_previous_context()

    def __str__(self):
        return (
            f"ModuleRecord({self.what}, {self.parent}, {self.original},"
            f" {self.replacement}) {self.caller_module}"
        )


def _apply_module_replacement(
    what,
    module,
    cassette,
    decorate,
    replace,
    module_record_list: List[ModuleRecord],
    add_revert_list: List,
) -> List[ModuleRecord]:
    """
    Internal method what finds inside module if the what string matches.
    If yes, apply the  replacement, otherwise return None

    :param what: What will be replaced
    :param module: the module where we try to find match for "what"
    :param cassette: Cassette instance
    :param decorate: decorator to be applied
    :param replace: replace function to be applied
    :param module_record_list: list of modules what was already applied
    :return: None or ModuleRecord
    """
    full_module_list = what.split(".")
    module_name = module.__name__
    output_record_list: List[ModuleRecord] = list()
    # avoid to deep dive into
    # if not matched, try to find just part, if not imported as full path
    for depth, _ in enumerate(full_module_list):
        original_obj = module
        parent_obj = None
        # rest of list has to match object path
        for module_path in full_module_list[depth:]:
            parent_obj = original_obj
            try:
                original_obj = getattr(original_obj, module_path)
            except AttributeError:
                # this is used also for indication that match of path passed
                # in case it match partial path, make it None
                parent_obj = None
                break
            original_obj_text = (
                original_obj.__name__
                if hasattr(original_obj, "__name__")
                else str(original_obj)
            )
            logger.debug(
                f"\tmodule {module_name} -> {module_path} in "
                f"{original_obj_text} ({full_module_list[depth:]})"
            )
        # continue if parent is empty or module of function did not match
        # path inside what avoid replace something else with same path
        # eg, you define what "re.search", and it matches "search" as
        # part of your module but it does not come from re module

        try:
            # TODO: this part should be improved, theoretically
            #  join(full_module_list).startswith( original_obj.__module__)
            #  could lead to issue, that it matches another module path
            if not parent_obj:
                continue
            if original_obj.__module__ and not ".".join(full_module_list).startswith(
                original_obj.__module__
            ):
                logger.debug(
                    f"SIMILAR MATCH module {module_name} "
                    f"in {parent_obj.__name__} -> {original_obj.__name__} "
                    f"from {original_obj.__module__} ({full_module_list[depth:]})"
                )
                continue
        except AttributeError as e:
            logger.debug(e)
            continue
        logger.info(
            f"MATCH module {module_name} "
            f"in {parent_obj.__name__} -> {original_obj.__name__} "
            f"from {original_obj.__module__} ({full_module_list[depth:]})"
        )
        if replace is not None:
            if isinstance(replace, CassetteExecution):
                new_function = replace.function
                replace.cassette = cassette
                replace.obj_cls.set_cassette(cassette)
            else:
                new_function = replace
            replacement = new_function
        elif decorate is not None:
            if not isinstance(decorate, list):
                new_function = [decorate]
            else:
                new_function = decorate
            replacement = original_obj
            for item in new_function:
                if isinstance(item, CassetteExecution):
                    item.cassette = cassette
                    item.obj_cls.set_cassette(cassette)
                    replacement = item.function(replacement)
                else:
                    replacement = item(replacement)
        else:
            continue
        # check if already replaced, then continue
        if original_obj in [x.replacement for x in module_record_list]:
            logger.info(f"\talready replaced {what} in {module_name}")
            continue
        # TODO: have to try to investigate how to do multiple replacements of same
        #  change the module string in replacements, to be able to do multiple
        #  replacements this is tricky and may be confusing, but also make it
        #  clear what has to be replaced
        #  replacement.__module__ = original_obj.__module__
        setattr(
            parent_obj,
            original_obj.__name__,
            replacement,
        )
        fn_str = (
            new_function.__name__
            if not isinstance(new_function, list)
            else new_function
        )
        logger.info(
            f"\tREPLACES {what} in {module_name}"
            f" by function {fn_str} {original_obj}"
        )
        output_record_list.append(
            ModuleRecord(
                what=what,
                parent=parent_obj,
                original=original_obj,
                replacement=replacement,
                add_revert_list=add_revert_list,
            )
        )
    return output_record_list


def _parse_and_replace_sys_modules(
    what: str,
    cassette: Cassette,
    decorate: Any = None,
    replace: Any = None,
    add_revert_list: Optional[List] = None,
) -> List[ModuleRecord]:
    """
    Internal fucntion what will check all sys.modules, and try to find there implementation of
    "what" and replace or decorate it by given value(s)
    """
    logger.info(f"\n++++++ SEARCH {what} decorator={decorate} replace={replace}")
    module_list: List[ModuleRecord] = []
    # go over all modules, and try to find match
    for module in sys.modules.copy().values():
        module_list += _apply_module_replacement(
            what=what,
            module=module,
            cassette=cassette,
            decorate=decorate,
            replace=replace,
            module_record_list=module_list,
            add_revert_list=add_revert_list,
        )
    return module_list


def change_storage_file(
    cassette: Cassette, func, args, storage_file: Optional[str] = None
):
    """
    Internal function that try to construct persistent data file based on various
    possibilities.

    :param cassette: Cassette instance to pass inside object to work with
    """
    if storage_file:
        cassette.storage_file = storage_file
    else:
        if len(args):
            try:
                cassette.storage_file = get_datafile_filename(args[0])
            except NameError:
                cassette.storage_file = get_datafile_filename(func)
        else:
            cassette.storage_file = get_datafile_filename(func)
    original_storage_file = cassette.storage_file
    return original_storage_file


def _revert_modules(module_list: List[ModuleRecord]):
    """
    Revert modules to original functions

    :param module_list: list of modules
    :return:
    """
    for module_info in module_list:
        setattr(
            module_info.parent,
            module_info.original.__name__,
            module_info.original,
        )
        if not module_info.add_revert_list:
            continue
        # revert also items on additional places if necessarry
        for item in module_info.add_revert_list:
            found = False
            main_key = item.split(".")[-1]
            module_key = ".".join(item.split(".")[:-1])
            for module_name, module in sys.modules.copy().items():
                # go through all modules and try to find if item is stored there
                if module_key.startswith(module_name):
                    parent_obj = module
                    try:
                        for key in module_key.lstrip(module_name)[1:].split("."):
                            parent_obj = getattr(parent_obj, key)
                        setattr(
                            parent_obj,
                            main_key,
                            module_info.original,
                        )
                        found = True
                    except AttributeError:
                        pass
            if not found:
                raise AttributeError(
                    "You've want to revert item, but not found anywhere inside loaded modules"
                )


def make_generic(_decorator=None):
    """
    Decorator for decorators to make them applicable both on classes and methods/functions.

    When the decorated decorator is applied on class, it will be applied on methods.
    By default, it applies the decorator to all `test.*` method;
    use `regexp_method_pattern` argument to change it.
    It also triggers `cassette_setup`/`cassette_teardown` method before/after method execution
    to be able to manipulate cassette in a method shared between all test cases.
    (We do not have access to cassette from regular setUp/tearDown method.)


    Decorator make_generic can be used both with and without parenthesis:

    Example:

    @make_generic
    def my_decorator(fce):
        return fce

    @my_decorator
    class ClassA:
        def method_b(self):
            print("Will be decorated.")
        def method_c(self):
            print("Will be decorated.")

    class ClassD:
        @my_decorator
        def method_e(self):
            print("Will be decorated.")


    Warning:
        If you want to use make_generic with decorator accepting key-word arguments,
        to be able to use the final decorator without arguments or without parenthesis,
        use the following trick to correctly apply all decorators:

        @make_generic
        def add_something(_func=None, add=1):
            def decorator_to_return(fce):
                def fce_to_return():
                    return fce() + add

                return fce_to_return

            if _func is None:
                # @add_something()
                return decorator_to_return
            else:
                # @add_something
                return decorator_to_return(_func)

    """

    def make_generic_decorator_cover(decorator):
        def make_generic_decorator(*args, regexp_method_pattern=None, **kwargs):

            is_direct_call = (
                not kwargs
                and not regexp_method_pattern
                and len(args) == 1
                and isinstance(args[0], (type, Callable))
            )

            if not is_direct_call and (args or kwargs):
                # decorator with arguments
                decorate = decorator(*args, **kwargs)
            else:
                # decorator without arguments
                decorate = decorator

            def decorated_decorator(function_or_class):
                if isinstance(function_or_class, type):
                    return apply_decorator_to_all_methods(
                        decorate, regexp_method_pattern=regexp_method_pattern
                    )(function_or_class)
                else:
                    return decorate(function_or_class)

            # To support syntax with and without parentheses
            # for the decorated decorator.
            if is_direct_call:
                # call without parenthesis, the target is in the arguments
                return decorated_decorator(args[0])
            else:
                # call with parenthesis, real target will be applied later
                return decorated_decorator

        return make_generic_decorator

    # To support syntax with and without parentheses of the make_generic itself.
    if _decorator is None:
        # @make_generic()
        return make_generic_decorator_cover
    else:
        # @make_generic
        return make_generic_decorator_cover(_decorator)


@make_generic
def replace(
    what: str,
    cassette: Optional[Cassette] = None,
    decorate: Optional[Union[List[Callable], Callable]] = None,
    replace: Optional[Callable] = None,
    storage_keys_strategy=None,
):
    """
    Decorator what helps you to replace/decorate functions/methods inside any already
    imported module. It uses what as identifier what you want to replace, then you can
    define if it will be decorated or replaced by given function.

    Be aware of several situations:
      * This will not work if there are dynamic imports inside code execution
        * Workaround: import the module as part of your test code fist
      * When you try to decorate/replace method what is not defined (e.g. via using __getattr__)
        * Workaround: try it. maybe it will work as expected.
        Import the module and create there mock the method, or replace it with inspired by
        __getattr__ codebase

    Example usage to decorate request method of requests module:
    @replace_module_match(what="requests.sessions.Session.request",
                          decorate=RequestResponseHandling.decorator(
                          item_list=["method", "url", "data"],
                          map_item_list={"url": remove_password_from_url})

    :param what: str - full path of function inside module
    :param decorate: function decorator what will be applied to what, could be also list of
                     decorators, to be able to apply more decorators on one function
                     eg. store files and store output
    :param replace: replace original function by given one
    :param cassette: Cassette instance to pass inside object to work with
    :param storage_keys_strategy: you can change key strategy for storing data
                                  default simple one avoid to store stack information
    """
    storage_keys_strategy = storage_keys_strategy or StorageKeysInspectSimple

    if decorate is None and replace is None:
        logger.info(f"Using default decorator for {what}")
        decorate = Guess.decorator_plain(cassette=cassette)
    elif decorate is not None and replace is not None:
        raise ValueError("right one from [decorate, replace] parameter has to be set.")

    def replace_decorator(func):
        func_cassette = (
            getattr(func, REQURE_CASSETTE_ATTRIBUTE_NAME)
            if hasattr(func, REQURE_CASSETTE_ATTRIBUTE_NAME)
            else None
        )
        cassette_int = cassette or func_cassette or Cassette()

        @functools.wraps(func)
        def _replaced_function(*args, **kwargs):
            # set storage if not set to default one, based on function name
            if cassette_int.storage_file is None:
                change_storage_file(cassette=cassette_int, func=func, args=args)
            cassette_int.data_miner.key_stategy_cls = storage_keys_strategy
            # ensure that directory structure exists already
            os.makedirs(os.path.dirname(cassette_int.storage_file), exist_ok=True)
            # Store values and their replacements for modules
            # to be able to _revert_modules changes back
            module_list = _parse_and_replace_sys_modules(
                what=what, cassette=cassette_int, decorate=decorate, replace=replace
            )
            try:
                # pass current cassette to underneath decorator and do not overwrite if set there
                if (
                    "cassette" in inspect.getfullargspec(func).annotations
                    and inspect.getfullargspec(func).annotations["cassette"] == Cassette
                    and "cassette" not in kwargs
                ):
                    kwargs["cassette"] = cassette_int
                # execute content
                output = func(*args, **kwargs)
            except Exception as e:
                raise (e)
            finally:
                # dump data to storage file
                cassette_int.dump()
                _revert_modules(module_list)
            return output

        setattr(_replaced_function, REQURE_CASSETTE_ATTRIBUTE_NAME, cassette_int)
        return _replaced_function

    return replace_decorator


@make_generic
def record(
    what: str,
    cassette: Optional[Cassette] = None,
):
    """
    Decorator which can be used to store calls of the function and
    and replay responses on the next run.

    Can be applied both on function/method and class.

    When the applied on class, it will be applied on methods.
    By default, it applies the decorator to all `test.*` method;
    use `regexp_method_pattern` argument to change it.
    It also triggers `cassette_setup`/`cassette_teardown` method before/after method execution
    to be able to manipulate cassette in a method shared between all test cases.
    (We do not have access to cassette from regular setUp/tearDown method.)

    If you want to have more options for the recording, use `replace` decorator instead.
    You can tweak the recording process, structure of the saved data
    and also provide your own way how to store side-effects of the function.

    If you want some pre-made record decorators of common use-cases
    (like for git, tempfile or requests) take a look to `requre.helpers.__init__`.
    They take care about the side-effects for you.

    Examples:

    @record(what="tests.data.special_requre_module.random_number")
    class RecordDecoratorForClass(unittest.TestCase):
        def test_random(self):
            random_number = tests.data.special_requre_module.random_number()
            self.assertEqual(random_number, 0.819349292484907)

    class RecordDecoratorForMethod(unittest.TestCase):
        @record(what="tests.data.special_requre_module.random_number")
        def test_random(self):
            random_number = tests.data.special_requre_module.random_number()
            self.assertEqual(random_number, 0.17583106733657616)


    :param what: str - full path of function inside module
    :param cassette: Cassette instance to pass inside object to work with
    """

    def _record_inner(func):
        return replace(what=what, cassette=cassette)(func)

    return _record_inner


@contextmanager
def recording(
    what: str,
    decorate: Optional[Union[List[Callable], Callable]] = None,
    replace: Optional[Callable] = None,
    storage_file: Optional[str] = None,
    storage_keys_strategy=StorageKeysInspectSimple,
):
    """
    Context manager which can be used to store calls of the function and
    and replay responses on the next run.

    :param what: str - full path of function inside module
    :param decorate: function decorator what will be applied to what, could be also list of
                     decorators, to be able to apply more decorators on one function
                     eg. store files and store output
    :param replace: replace original function by given one
    :param storage_file: path for storage file if you don't want to use default location
    :param storage_keys_strategy: you can change key strategy for storing data
                                  default simple one avoid to store stack information
    """
    cassette = Cassette()
    cassette.storage_file = storage_file
    cassette.data_miner.key_stategy_cls = storage_keys_strategy
    # ensure that directory structure exists already
    os.makedirs(os.path.dirname(cassette.storage_file), exist_ok=True)
    # use default decorator for context manager if not given.
    if decorate is None and replace is None:
        logger.info(f"Using default decorator for {what}")
        decorate = Guess.decorator_plain(cassette=cassette)
    elif decorate is not None and replace is not None:
        raise ValueError("right one from [decorate, replace] parameter has to be set.")
    # Store values and their replacements for modules to be able to _revert_modules changes back
    module_list = _parse_and_replace_sys_modules(
        what=what, cassette=cassette, decorate=decorate, replace=replace
    )
    try:
        yield cassette
    finally:
        # dump data to storage file
        cassette.dump()
        _revert_modules(module_list)


def cassette_setup_and_teardown_decorator(func):
    """
    Decorator that triggers `cassette_setup` method to be run before the test method.
    """
    if hasattr(func, REQURE_SETUP_APPLIED_ATTRIBUTE_NAME):
        return func

    func_cassette = (
        getattr(func, REQURE_CASSETTE_ATTRIBUTE_NAME)
        if hasattr(func, REQURE_CASSETTE_ATTRIBUTE_NAME)
        else None
    )
    cassette_int = func_cassette or Cassette()

    @functools.wraps(func)
    def cassette_setup_inner(self, *args, **kwargs):
        if hasattr(self, "cassette_setup"):
            self.cassette_setup(cassette=cassette_int)

        if (
            "cassette" in inspect.getfullargspec(func).annotations
            and inspect.getfullargspec(func).annotations["cassette"] == Cassette
            and "cassette" not in kwargs
        ):
            kwargs["cassette"] = cassette_int

        return_value = func(self, *args, **kwargs)

        if hasattr(self, "cassette_teardown"):
            self.cassette_teardown(cassette=cassette_int)

        return return_value

    setattr(cassette_setup_inner, REQURE_CASSETTE_ATTRIBUTE_NAME, cassette_int)
    setattr(cassette_setup_inner, REQURE_SETUP_APPLIED_ATTRIBUTE_NAME, True)
    return cassette_setup_inner


def apply_decorators_recursively_to_fn(decorator_list, func):
    if not decorator_list:
        return func
    recursive_result = apply_decorators_recursively_to_fn(decorator_list[:-1], func)
    return decorator_list[-1](recursive_result)


def apply_decorator_to_all_methods(*decorator: Callable, regexp_method_pattern=None):
    """
    This function works as class decorator and apply decorator to
    all matched methods via regexp (`test.*` by default),
    primary usage is to use it for unittest testcases.

    ref: https://stackoverflow.com/a/6307868

    Also triggers `cassette_setup`/`cassette_teardown` method before/after method execution
    to be able to manipulate cassette in a method shared between all test cases.
    (We do not have access to cassette from regular setUp/tearDown method.)
    """
    regexp_method_pattern = regexp_method_pattern or TEST_METHOD_REGEXP

    # append cassette setup and teardown if exist
    decorator += (cassette_setup_and_teardown_decorator,)

    def decorate(cls):
        for attr in cls.__dict__:
            if callable(getattr(cls, attr)) and re.match(regexp_method_pattern, attr):
                setattr(
                    cls,
                    attr,
                    apply_decorators_recursively_to_fn(
                        decorator_list=decorator, func=getattr(cls, attr)
                    ),
                )
        return cls

    return decorate


def replace_module_match_with_multiple_decorators(
    *decorators: Tuple[str, Callable],
    cassette: Optional[Cassette] = None,
):
    if not decorators:
        raise AttributeError("decorators parameter has to be defined")
    decorator_list = []
    for what, decorate in decorators:
        decorator_list.append(replace(what=what, decorate=decorate, cassette=cassette))

    def decorator(func):
        return apply_decorators_recursively_to_fn(decorator_list, func)

    return decorator



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import os

import tempfile
from requre.helpers.tempfile import TempFile, MkTemp, MkDTemp
from requre.utils import get_datafile_filename
from tests.testbase import BaseClass
from requre.online_replacing import replace


class TestTempFile(BaseClass):
    def setUp(self) -> None:
        super().setUp()
        self.cassette.storage_file = get_datafile_filename(self.id)
        TempFile.set_cassette(self.cassette)

    def testSimple(self):
        output = TempFile._mktemp()
        self.assertIn(
            f"/tmp/{os.path.basename(self.cassette.storage_file)}/static_tmp_1",
            output,
        )

    def testChangeFile(self):
        self.cassette.storage_file = str(self.cassette.storage_file) + ".x"
        output = TempFile._mktemp()
        self.assertIn(
            f"/tmp/{os.path.basename(self.cassette.storage_file)}/static_tmp_1",
            output,
        )
        output = TempFile._mktemp()
        self.assertIn(
            f"/tmp/{os.path.basename(self.cassette.storage_file)}/static_tmp_2",
            output,
        )
        self.cassette.storage_file = str(self.cassette.storage_file) + ".y"
        self.assertEqual(TempFile.counter, 2)
        output = TempFile._mktemp()
        self.assertEqual(TempFile.counter, 1)
        self.assertIn(
            f"/tmp/{os.path.basename(self.cassette.storage_file)}/static_tmp_1",
            output,
        )
        output = TempFile._mktemp()
        self.assertIn(
            f"/tmp/{os.path.basename(self.cassette.storage_file)}/static_tmp_2",
            output,
        )


@replace(what="tempfile.mktemp", decorate=MkTemp.decorator_plain())
def new_tempfile():
    return tempfile.mktemp()


@replace(what="tempfile.mkdtemp", decorate=MkDTemp.decorator_plain())
def new_tempdir():
    return tempfile.mkdtemp()


class TempFile_New(BaseClass):
    def test_tempfile(self):
        """When regeneration, tempfile will change, so change the expected output"""
        filename = new_tempfile()
        self.assertFalse(os.path.exists(filename))
        self.assertEqual(filename, "/tmp/tmp5hi99k_2")

    def test_tempdir(self):
        """When regeneration, tempdir will change, so change the expected output"""
        filename = new_tempdir()
        self.assertTrue(os.path.exists(filename))
        self.assertEqual(filename, "/tmp/tmpy7vcma7c")



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import builtins

VERSION_REQURE_FILE = 3
ENV_STORAGE_FILE = "RESPONSE_FILE"
ENV_REPLACEMENT_FILE = "REPLACEMENT_FILE"
ENV_REPLACEMENT_NAME = "REPLACEMENT_VAR"
ENV_DEBUG = "DEBUG"
ENV_APPLY_LATENCY = "LATENCY"
ENV_REQURE_STORAGE_MODE = "REQURE_MODE"
REPLACE_DEFAULT_KEY = "FILTERS"
METATADA_KEY = "_requre"
KEY_MINIMAL_MATCH = 2
RELATIVE_TEST_DATA_DIRECTORY = "test_data"
DEFAULT_SUFIX = "yaml"
REQURE_CASSETTE_ATTRIBUTE_NAME = "_requre_cassette"
REQURE_SETUP_APPLIED_ATTRIBUTE_NAME = "_requre_cassette_setup_applied"
TEST_METHOD_REGEXP = "test.*"

DEFAULT_IMPORT_FUNCTION = builtins.__import__



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import os
import logging
from typing import Union, Any, Dict, Optional, List
from .constants import KEY_MINIMAL_MATCH, METATADA_KEY
from .cassette import DataMiner, DataStructure, DataTypes
from glob import glob
import tarfile
import tempfile
import shutil
from subprocess import check_output, CalledProcessError
import hashlib
from _hashlib import HASH as Hash
from pathlib import Path

logger = logging.getLogger(__name__)


class DictProcessing:
    def __init__(self, requre_dict: dict):
        self.requre_dict = requre_dict

    def match(self, selector: list, internal_object: Union[dict, list, None] = None):
        if internal_object is None:
            internal_object = self.requre_dict
        if len(selector) == 0:
            logger.debug("all selectors matched")
            yield internal_object
            # add return here, to avoid multiple returns
            return
        if isinstance(internal_object, dict):
            for k, v in internal_object.items():
                if v is None:
                    return
                if selector and selector[0] == k:
                    logger.debug(f"selector {k} matched")
                    yield from self.match(selector=selector[1:], internal_object=v)
                else:
                    yield from self.match(selector=selector, internal_object=v)
        elif isinstance(internal_object, list):
            for list_item in internal_object:
                if list_item is None:
                    return
                yield from self.match(selector=selector, internal_object=list_item)
        else:
            return

    @staticmethod
    def replace(obj: Any, key: Any, value: Any) -> None:
        if isinstance(obj, dict):
            for k, v in obj.items():
                if k == key:
                    logger.debug(f"replacing: {obj[key]} by {value}")
                    obj[key] = value
                else:
                    DictProcessing.replace(obj=v, key=key, value=value)
        if isinstance(obj, list):
            for item in obj:
                DictProcessing.replace(obj=item, key=key, value=value)

    @staticmethod
    def minimal_match(dict_obj: Dict, metadata: Dict):

        tmp_dict = dict_obj
        first_item: Dict = {}
        key_name = DataTypes.__name__
        ds = DataTypes.List
        if key_name in metadata:
            ds = DataTypes(metadata.get(key_name))
        logger.debug(f"Use datatype: {ds}")
        for cntr in range(KEY_MINIMAL_MATCH):
            if not isinstance(tmp_dict, dict) or len(tmp_dict.keys()) != 1:
                return False
            key = list(tmp_dict.keys())[0]
            value = tmp_dict[key]
            tmp_dict = value
        if ds == DataTypes.DictWithList:
            if isinstance(tmp_dict, dict):
                tmp_first_item = tmp_dict[list(tmp_dict.keys())[0]]
                if isinstance(tmp_first_item, list):
                    first_item = tmp_first_item[0]
        if ds == DataTypes.Dict:
            if isinstance(tmp_dict, dict):
                first_item = tmp_dict[list(tmp_dict.keys())[0]]
        if ds == DataTypes.List:
            if isinstance(tmp_dict, list):
                first_item = tmp_dict[0]
        if ds == DataTypes.Value:
            if isinstance(tmp_dict, dict):
                first_item = tmp_dict
        if isinstance(first_item, dict) and DataMiner().LATENCY_KEY in first_item.get(
            DataStructure.METADATA_KEY, {}
        ):
            logger.debug(
                f"Minimal key path ({KEY_MINIMAL_MATCH}) matched for {first_item.keys()}"
            )
            return True
        return False

    def simplify(
        self, internal_object: Optional[Dict] = None, ignore_list: Optional[List] = None
    ):
        ignore_list = ignore_list or []
        internal_object = (
            self.requre_dict if internal_object is None else internal_object
        )
        if isinstance(internal_object, dict):
            if len(internal_object.keys()) == 1:
                key = list(internal_object.keys())[0]
                if key in [METATADA_KEY] + ignore_list:
                    return
                if self.minimal_match(internal_object, self.requre_dict[METATADA_KEY]):
                    return
                if isinstance(internal_object[key], dict):
                    value = internal_object.pop(key)
                    print(f"Removing key: {key}  -> {list(value.keys())}")
                    for k, v in value.items():
                        internal_object[k] = v
                        self.simplify(
                            internal_object=internal_object, ignore_list=ignore_list
                        )
            else:
                for v in internal_object.values():
                    self.simplify(internal_object=v, ignore_list=ignore_list)


class TarFilesSimilarity:
    def __init__(self, path, hash_function=None):
        self._mapping_table = None
        self.path = path
        self.hash_function = hash_function

    @property
    def mapping_table(self):
        if self._mapping_table is None:
            self._mapping_table = dict()
            for tar_file in self.list_files(self.path):
                current_hash = self.untar_and_return_hash(tar_file, self.hash_function)
                self._mapping_table[tar_file] = current_hash
        return self._mapping_table

    @staticmethod
    def md5_update_from_file(filename: Union[str, Path], hash: Hash) -> Hash:
        # https://stackoverflow.com/questions/24937495/how-can-i-calculate-a-hash-for-a-filesystem-directory-using-python
        assert Path(filename).is_file()
        with open(str(filename), "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash.update(chunk)
        return hash

    @staticmethod
    def md5_file(filename: Union[str, Path]) -> str:
        return str(
            TarFilesSimilarity.md5_update_from_file(filename, hashlib.md5()).hexdigest()
        )

    @staticmethod
    def md5_update_from_dir(directory: Union[str, Path], hash: Hash) -> Hash:
        assert Path(directory).is_dir()
        for path in sorted(Path(directory).iterdir(), key=lambda p: str(p).lower()):
            hash.update(path.name.encode())
            if path.is_file():
                hash = TarFilesSimilarity.md5_update_from_file(path, hash)
            elif path.is_dir():
                hash = TarFilesSimilarity.md5_update_from_dir(path, hash)
        return hash

    @staticmethod
    def md5_dir(directory: Union[str, Path]) -> str:
        return str(
            TarFilesSimilarity.md5_update_from_dir(directory, hashlib.md5()).hexdigest()
        )

    @staticmethod
    def list_files(path, pattern="*.tar.xz"):
        return glob(f"{path}/**/{pattern}", recursive=True)

    @staticmethod
    def hash_fn(content):
        if os.path.isdir(content):
            try:
                os.chdir(content)
                return "git-sha", check_output(["git", "rev-parse", "HEAD"])
            except (FileNotFoundError, CalledProcessError):
                return "md5-dir", TarFilesSimilarity.md5_dir(content)
        else:
            return "md5-file", TarFilesSimilarity.md5_file(content)

    @staticmethod
    def untar_and_return_hash(tar_archive_path, hash_funciton):
        hash_funciton = hash_funciton or TarFilesSimilarity.hash_fn
        tempdir = tempfile.mkdtemp()
        currentdir = os.getcwd()
        try:
            os.chdir(tempdir)
            tar_file = tarfile.open(tar_archive_path)
            tar_file.extractall()
            files = os.listdir(".")
            if len(files) != 1:
                logger.debug("tar not packed by requre")
                return None
            content = files[0]
            return hash_funciton(content)
        finally:
            os.chdir(currentdir)
            shutil.rmtree(tempdir, ignore_errors=True)

    def find_same(self):
        item_dict = {}
        for k, v in self.mapping_table.items():
            if v not in item_dict:
                item_dict[v] = [k]
            else:
                item_dict[v].append(k)
        return item_dict

    def symlink_same_files(self):
        for k, v in self.find_same().items():
            if k is None:
                logger.debug(f"items not created by requre: {v}")
            if len(v) == 1:
                logger.debug(f"no duplication found: {v}")
            else:
                main_file = v[0]
                relative_main = main_file.replace(self.path, "")
                rest_files = v[1:]
                for file in rest_files:
                    if os.path.islink(file):
                        logger.debug(
                            f"{file} already symlinked to {os.path.realpath(file)}"
                        )
                        continue
                    relative_file = file.replace(self.path, "")
                    dir_deep = len(
                        os.path.dirname(relative_file)
                        .strip(os.path.sep)
                        .split(os.path.sep)
                    )
                    main_relative_file = f"..{os.path.sep}" * dir_deep + relative_main
                    os.remove(file)
                    os.symlink(main_relative_file, file)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from requre.import_system import UpgradeImportSystem
from requre.helpers.tempfile import TempFile

special = UpgradeImportSystem().decorate("tempfile.mktemp", TempFile.mktemp())



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import functools
import inspect
import logging
import pickle
import warnings
from typing import Optional, Callable, Any, List, Dict

from requre.storage import PersistentObjectStorage
from requre.cassette import (
    original_time,
    StorageKeysInspectFull,
    Cassette,
    CassetteExecution,
)


logger = logging.getLogger(__name__)


class ObjectStorage:
    """
    Generic object API for objects for persistent storage.
    Use it as parent class for your custom object handling

    This class use pickle module for object serialization,
    it is generic, but may lead to some issues
    in case object is not serializable well. Use this class very carefully.
    """

    _cassette: Cassette = None
    __response_keys: list = list()
    object_type = object
    DUPLICATION_KEY = "requre.objects"
    stack_internal_check = True

    def __init__(
        self,
        store_keys: list,
        cassette: Cassette,
        storage_object_kwargs: Optional[dict] = None,
    ) -> None:
        self.store_keys = store_keys
        if cassette:
            self.set_cassette(cassette)
        self.store_keys = store_keys
        self.storage_object_kwargs = storage_object_kwargs or {}

    @classmethod
    def get_cassette(cls):
        """
        Internal method to return proper cassette, it is important after rewriting of singletons
        Cassette were not passed to object executor
        :return: Cassette instance
        """
        if cls._cassette:
            return cls._cassette
        else:
            return PersistentObjectStorage().cassette

    @classmethod
    def set_cassette(cls, value):
        cls._cassette = value

    @classmethod
    def get_base_keys(cls, func: Callable) -> List[Any]:
        function = func.function if isinstance(func, CassetteExecution) else func
        return cls.get_cassette().data_miner.key_stategy_cls.get_base_keys(function)

    @classmethod
    def execute(
        cls,
        keys: list,
        func: Callable,
        *args,
        storage_object_kwargs=None,
        cassette: Cassette,
        **kwargs,
    ) -> Any:
        """
        Class method to store or read object from persistent storage
        with explicit set of *args, **kwargs parameters to use as keys
        :param keys: list of keys used as parameter
        :param func: original function
        :param args: parameters of original function
        :param storage_object_kwargs: forwarded to the storage object
        :param cassette: Cassette instance to pass inside object to work with
        :param kwargs: parameters of original function
        :return: CassetteExecution class with function and cassette instance
        """
        storage_object_kwargs = storage_object_kwargs or {}
        object_storage = cls(
            store_keys=keys, cassette=cassette, **storage_object_kwargs
        )

        if object_storage.get_cassette().do_store(keys):
            time_before = original_time()
            func_exposed = (
                func.function if isinstance(func, CassetteExecution) else func
            )
            response = func_exposed(*args, **kwargs)

            time_after = original_time()
            call_stack = StorageKeysInspectFull.get_base_keys(func_exposed)
            # do not store data of fuction what will be stored by upper decodator
            if cls.stack_internal_check and call_stack.count(cls.DUPLICATION_KEY) > 1:
                return response
            metadata: Dict = {
                cassette.data_miner.LATENCY_KEY: time_after - time_before,
                cassette.data_miner.METADATA_CALLER_LIST: call_stack,
            }
            if cassette.data_miner.store_arg_debug_metadata:
                args_clean = [f"'{x}'" if isinstance(x, str) else str(x) for x in args]
                kwargs_clean = [
                    f"""{k}={f"'{v}'" if isinstance(v, str) else str(v)}"""
                    for k, v in kwargs.items()
                ]
                caller = (
                    f"{func_exposed.__name__}({', '.join(args_clean + kwargs_clean)})"
                )
                metadata[cassette.data_miner.METADATA_ARG_DEBUG_KEY] = caller
            object_storage.write(response, metadata)
            logger.debug(f"WRITE Keys: {keys} -> {response}")
            return response

        else:
            response = object_storage.read()
            logger.debug(f"READ  Keys: {keys} -> {response}")
            return response

    @classmethod
    def execute_all_keys(
        cls,
        func: Callable,
        *args,
        storage_object_kwargs=None,
        cassette: Cassette,
        **kwargs,
    ):
        """
        Class method what does same as execute, but use all *args, **kwargs as keys
        :param func: original function
        :param args: parameters of original function
        :param storage_object_kwargs: forwarded to the storage object
        :param kwargs: parameters of original function
        :param cassette: Cassette instance to pass inside object to work with
        :return: CassetteExecution class with function and cassette instance
        """
        func_exposed = func.function if isinstance(func, CassetteExecution) else func
        keys = (
            cls.get_base_keys(func_exposed)
            + [x for x in args if isinstance(int, str)]
            + [f"{k}:{v}" for k, v in kwargs.items()]
        )
        return cls.execute(
            keys,
            func,
            *args,
            storage_object_kwargs=storage_object_kwargs,
            cassette=cassette,
            **kwargs,
        )

    @classmethod
    def execute_plain(
        cls,
        func: Callable,
        *args,
        storage_object_kwargs=None,
        cassette: Cassette,
        **kwargs,
    ):
        """
        Class method what does same as execute, but use just name of module and function as name

        :param func: original function
        :param args: parameters of original function
        :param storage_object_kwargs: forwarded to the storage object
        :param kwargs: parameters of original function
        :param cassette: Cassette instance to pass inside object to work with
        :return: CassetteExecution class with function and cassette instance
        """
        func_exposed = func.function if isinstance(func, CassetteExecution) else func
        keys = cls.get_base_keys(func_exposed)
        return cls.execute(
            keys,
            func,
            *args,
            storage_object_kwargs=storage_object_kwargs,
            cassette=cassette,
            **kwargs,
        )

    @classmethod
    def decorator_all_keys(
        cls,
        storage_object_kwargs=None,
        cassette: Cassette = None,
    ) -> Any:
        """
        Class method for what should be used as decorator of import replacing system
        This use all arguments of function as keys

        :param func: Callable object
        :param storage_object_kwargs: forwarded to the storage object
        :param cassette: Cassette instance to pass inside object to work with
        :return: CassetteExecution class with function and cassette instance
        """

        def internal(func):
            @functools.wraps(func)
            def internal_internal(*args, **kwargs):
                return cls.decorator(
                    item_list=list(range(len(args))) + list(kwargs.keys()),
                    cassette=cassette,
                )(func)(*args, **kwargs)

            return internal_internal

        return internal

    @classmethod
    def decorator(
        cls,
        *,
        item_list: list,
        map_function_to_item=None,
        storage_object_kwargs=None,
        cassette: Cassette = None,
    ) -> Any:
        """
        Class method for what should be used as decorator of import replacing system
        This use list of selection of *args or **kwargs as arguments of function as keys

        :param item_list: list of values of *args nums,  **kwargs names to use as keys
        :param map_function_to_item: dict of function to apply to keys before storing
                                  (have to be listed in item_list)
        :param storage_object_kwargs: forwarded to the storage object
        :param cassette: Cassette instance to pass inside object to work with
        :return: CassetteExecution class with function and cassette instance
        """

        map_function_to_item = map_function_to_item or {}
        casex = CassetteExecution()
        casex.cassette = cassette or cls.get_cassette()
        casex.obj_cls = cls

        def internal(func: Callable):
            @functools.wraps(func)
            def internal_internal(*args, **kwargs):
                keys = cls.get_base_keys(func)
                # get all possible arguments of passed function
                try:
                    arg_keys = inspect.getfullargspec(func)[0]
                except TypeError:
                    arg_keys = []
                for param_name in item_list:
                    # if you pass int as an agrument, it forces to use args
                    if isinstance(param_name, int):
                        key = args[param_name]
                    # try to look into kwargs if item is there
                    elif param_name in kwargs:
                        key = kwargs[param_name]
                    #  translate param name to positional argument
                    else:
                        # out of index check. This is bad but possible use case
                        # raise warning and continue
                        if len(args) <= arg_keys.index(param_name):
                            warnings.warn(
                                f"You've defined keys: {item_list} but '{param_name}' is not part"
                                f" of args:{args} and kwargs:{kwargs},"
                                f" original function and args: {func.__name__}({arg_keys})"
                            )
                            # but add there None as key, to not spoil dictionary deep
                            key = None
                        else:
                            key = args[arg_keys.index(param_name)]
                    if param_name not in map_function_to_item:
                        keys.append(key)
                    else:
                        keys.append(map_function_to_item[param_name](key))
                return cls.execute(
                    keys,
                    func,
                    *args,
                    storage_object_kwargs=storage_object_kwargs,
                    cassette=casex.cassette,
                    **kwargs,
                )

            return internal_internal

        casex.function = internal
        return casex

    @classmethod
    def decorator_plain(
        cls,
        *,
        cassette: Cassette = None,
        storage_object_kwargs=None,
    ) -> Any:
        return cls.decorator(
            item_list=[],
            cassette=cassette,
            storage_object_kwargs=storage_object_kwargs,
        )

    def write(self, obj: Any, metadata: Optional[Dict] = None) -> Any:
        """
        Write the object representation to storage
        Internally it will use self.to_serializable
        method to get serializable object representation

        :param obj: some object
        :param metadata: store metedata to object
        :return: same obj
        """
        self.get_cassette().store(
            self.store_keys, self.to_serializable(obj), metadata=metadata
        )
        return obj

    def read(self):
        """
        Crete object representation of serialized data in persistent storage
        Internally it will use self.from_serializable method transform object

        :return: proper object
        """
        data = self.get_cassette()[self.store_keys]
        obj = self.from_serializable(data)
        return obj

    def to_serializable(self, obj: Any) -> Any:
        """
        Internal method for object serialization

        :param obj: some object
        :return: Yaml serializable object
        """
        output = pickle.dumps(obj)
        return output

    def from_serializable(self, data: Any) -> Any:
        """
        Internal method for object de-serialization

        :param data: Yaml serializable object
        :return: some object
        """
        output = pickle.loads(data)
        return output



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest

from tests.testbase import network_connection_available

pytest_plugins = ["requre.pytest_fixtures"]


@pytest.fixture(autouse=True)
def record_requests_fixture_autouse(record_requests_fixture):
    yield record_requests_fixture


@pytest.mark.skipif(not network_connection_available(), reason="No network connection")
def test_record_requests_fixture():
    import requests

    requests.get("https://google.com")


@pytest.mark.skipif(not network_connection_available(), reason="No network connection")
def test_record_requests_fixture_different_call():
    import requests

    requests.get("https://fedoraproject.org")



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT


import datetime
import json
from contextlib import contextmanager
from io import BytesIO
from typing import Any, Dict, List, Optional
from urllib.parse import urlparse

from requests.models import PreparedRequest, Request, Response
from requests.structures import CaseInsensitiveDict

from requre.cassette import Cassette
from requre.objects import ObjectStorage
from requre.record_and_replace import make_generic, recording, replace


def remove_password_from_url(url):
    urlobject = urlparse(url)
    if urlobject.password:
        return urlobject._replace(
            netloc="{}:{}@{}".format(urlobject.username, "???", urlobject.hostname)
        ).geturl()
    else:
        return url


class RequestResponseHandling(ObjectStorage):
    __response_keys = ["status_code", "encoding", "reason"]
    __ignored = ["cookies"]
    __response_keys_special = ["raw", "_next", "headers", "elapsed", "_content"]
    __store_indicator = "__store_indicator"
    __implicit_encoding = "UTF-8"

    def __init__(
        self,
        store_keys: list,
        cassette: Optional[Cassette] = None,
        response_headers_to_drop=None,
    ) -> None:
        # replace request if given as key and use prettier url
        for index, key in enumerate(store_keys):
            if isinstance(key, (Request, PreparedRequest)):
                store_keys[index] = remove_password_from_url(key.url)
                store_keys.insert(index, key.method)
        super().__init__(store_keys, cassette=cassette)
        self.response_headers_to_drop = response_headers_to_drop or []

    def write(self, response: Response, metadata: Optional[Dict] = None) -> Response:
        super().write(response, metadata)
        # TODO: disabled for now, improve next handling if we find it makes sense
        # if getattr(response, "next"):
        #    self.write(getattr(response, "next"))
        return response

    def read(self):
        data = super().read()
        # TODO: disabled for now, improve next handling if we find it makes sense
        # if getattr(data, "next"):
        #    data._next = self.read()
        return data

    def to_serializable(self, response: Response) -> Any:
        output = dict()
        for key in self.__response_keys:
            output[key] = getattr(response, key)
        for key in self.__response_keys_special:
            if key == "raw":
                binary_data = response.raw.read()
                output[key] = binary_data
                # replay it back to raw
                response.raw = BytesIO(binary_data)
            if key == "headers":
                headers_dict = dict(response.headers)
                for header in self.response_headers_to_drop:
                    if header in headers_dict:
                        headers_dict[header] = None
                output[key] = headers_dict
            if key == "elapsed":
                output[key] = response.elapsed.total_seconds()
            if key == "_content":
                what_store = response._content  # type: ignore
                encoding = response.encoding or self.__implicit_encoding
                try:
                    what_store = what_store.decode(encoding)  # type: ignore
                    try:
                        what_store = json.loads(what_store)
                        indicator = 2
                    except json.decoder.JSONDecodeError:
                        indicator = 1
                except (ValueError, AttributeError):
                    indicator = 0
                output[key] = what_store
                output[self.__store_indicator] = indicator
            if key == "_next":
                output[key] = None
                if getattr(response, "next") is not None:
                    output[key] = self.store_keys
        return output

    def from_serializable(self, data: Any) -> Response:
        response = Response()
        for key in self.__response_keys:
            setattr(response, key, data[key])
        for key in self.__response_keys_special:
            if key == "raw":
                response.raw = BytesIO(data[key])
            if key == "headers":
                response.headers = CaseInsensitiveDict(data[key])
            if key == "elapsed":
                response.elapsed = datetime.timedelta(seconds=data[key])
            if key == "_content":
                encoding = response.encoding or self.__implicit_encoding
                indicator = data[self.__store_indicator]
                if indicator == 0:
                    what_store = data[key]
                elif indicator == 1:
                    what_store = data[key].encode(encoding)
                elif indicator == 2:
                    what_store = json.dumps(data[key])
                    what_store = what_store.encode(encoding)
                response._content = what_store  # type: ignore
            if key == "_next":
                setattr(response, "_next", data[key])
        return response

    @classmethod
    def decorator_all_keys(
        cls,
        storage_object_kwargs=None,
        cassette: Cassette = None,
        response_headers_to_drop=None,
    ) -> Any:
        """
        Class method for what should be used as decorator of import replacing system
        This use all arguments of function as keys

        :param func: Callable object
        :param storage_object_kwargs: forwarded to the storage object
        :param response_headers_to_drop: list of header names we don't want to save with response
                                            (Will be replaced to `None`.)
        :param cassette: Cassette instance to pass inside object to work with
        :return: CassetteExecution class with function and cassette instance
        """
        storage_object_kwargs = storage_object_kwargs or {}
        if response_headers_to_drop:
            storage_object_kwargs["response_headers_to_drop"] = response_headers_to_drop
        return super().decorator_all_keys(
            storage_object_kwargs,
            cassette=cassette,
        )

    @classmethod
    def decorator(
        cls,
        *,
        item_list: list,
        map_function_to_item=None,
        storage_object_kwargs=None,
        cassette: Cassette = None,
        response_headers_to_drop=None,
    ) -> Any:
        """
        Class method for what should be used as decorator of import replacing system
        This use list of selection of *args or **kwargs as arguments of function as keys

        :param item_list: list of values of *args nums,  **kwargs names to use as keys
        :param map_function_to_item: dict of function to apply to keys before storing
                                  (have to be listed in item_list)
        :param storage_object_kwargs: forwarded to the storage object
        :param response_headers_to_drop: list of header names we don't want to save with response
                                        (Will be replaced to `None`.)
        :param cassette: Cassette instance to pass inside object to work with
        :return: CassetteExecution class with function and cassette instance
        """
        storage_object_kwargs = storage_object_kwargs or {}
        if response_headers_to_drop:
            storage_object_kwargs["response_headers_to_drop"] = response_headers_to_drop
        return super().decorator(
            item_list=item_list,
            map_function_to_item=map_function_to_item,
            storage_object_kwargs=storage_object_kwargs,
            cassette=cassette,
        )

    @classmethod
    def decorator_plain(
        cls,
        storage_object_kwargs=None,
        cassette: Cassette = None,
        response_headers_to_drop=None,
    ) -> Any:
        """
        Class method for what should be used as decorator of import replacing system
        This use no arguments of function as keys

        :param func: Callable object
        :param storage_object_kwargs: forwarded to the storage object
        :param response_headers_to_drop: list of header names we don't want to save with response
                                          (Will be replaced to `None`.)
        :param cassette: Cassette instance to pass inside object to work with
        :return: CassetteExecution class with function and cassette instance
        """
        storage_object_kwargs = storage_object_kwargs or {}
        if response_headers_to_drop:
            storage_object_kwargs["response_headers_to_drop"] = response_headers_to_drop
        return super().decorator_plain(
            storage_object_kwargs=storage_object_kwargs,
            cassette=cassette,
        )


@make_generic
def record_requests(
    _func=None,
    response_headers_to_drop: Optional[List[str]] = None,
    cassette: Optional[Cassette] = None,
):
    """
    Decorator which can be used to store all requests to a file
    and replay responses on the next run.

    - The matching is based on `url`.
    - Removes tokens from the url when saving if needed.

    Can be used with or without parenthesis.

    :param _func: can be used to decorate function (with, or without parenthesis).
    :param response_headers_to_drop: list of header names we don't want to save with response
                                        (Will be replaced to `None`.)
    :param storage_file: str - storage file to be passed to cassette instance if given,
                               else it creates new instance
    :param cassette: Cassette instance to pass inside object to work with
    """

    response_headers_to_drop = response_headers_to_drop or []
    replace_decorator = replace(
        what="requests.sessions.Session.send",
        cassette=cassette,
        decorate=RequestResponseHandling.decorator(
            item_list=[1],
            response_headers_to_drop=response_headers_to_drop,
            cassette=cassette,
        ),
    )

    if _func is not None:
        return replace_decorator(_func)
    else:
        return replace_decorator


@contextmanager
def recording_requests(
    response_headers_to_drop: Optional[List[str]] = None, storage_file=None
):
    """
    Context manager which can be used to store all requests to a file
    and replay responses on the next run.

    - The matching is based on `url`.
    - Removes tokens from the url when saving if needed.

    :param _func: can be used to decorate function (with, or without parenthesis).
    :param response_headers_to_drop: list of header names we don't want to save with response
                                        (Will be replaced to `None`.)
    :param storage_file: file for reading and writing data in storage_object
    """
    with recording(
        what="requests.sessions.Session.send",
        decorate=RequestResponseHandling.decorator(
            item_list=[1],
            response_headers_to_drop=response_headers_to_drop,
        ),
        storage_file=storage_file,
    ) as cassette:
        yield cassette



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import inspect
import logging
import shlex
import subprocess
from enum import Enum
from pathlib import Path

from _pytest.python import Function

from requre.constants import RELATIVE_TEST_DATA_DIRECTORY, DEFAULT_SUFIX
from requre.exceptions import PersistentStorageException

logger = logging.getLogger(__name__)


class StorageMode(Enum):
    default = 0
    read = 1
    write = 2
    append = 3


def run_command(cmd, error_message=None, cwd=None, fail=True, output=False):
    """
    subprocess wrapper, copied from packit, for higher level handling of executiong commands,
    :param cmd:
    :param error_message:
    :param cwd:
    :param fail:
    :param output:
    :return:
    """
    if not isinstance(cmd, list):
        cmd = shlex.split(cmd)

    logger.debug("cmd = '%s'", " ".join(cmd))

    cwd = cwd or str(Path.cwd())
    error_message = error_message or f"Command {cmd} failed."

    shell = subprocess.run(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=False,
        cwd=cwd,
        universal_newlines=True,
    )

    if not output:
        # output is returned, let the caller process it
        logger.debug("%s", shell.stdout)
    stderr = shell.stderr.strip()
    if stderr:
        logger.error("%s", shell.stderr)

    if shell.returncode != 0:
        logger.error("Command %s failed", shell.args)
        logger.error("%s", error_message)
        if fail:
            raise PersistentStorageException(
                f"Command {shell.args!r} failed: {error_message}"
            )
        success = False
    else:
        success = True

    if not output:
        return success
    return shell.stdout


class Replacement:
    def __init__(self, name, key, parent, one_filter, replacement) -> None:
        self.name = name
        self.key = key
        self.parent = parent
        self.filter = one_filter
        self.replacement = replacement


def get_module_of_previous_context():
    current_ctx = inspect.currentframe()
    while True:
        current_ctx = current_ctx.f_back
        frameinfo_args = (current_ctx,) + inspect.getframeinfo(current_ctx, 1)
        frameinfo = inspect.FrameInfo(*frameinfo_args)
        module = inspect.getmodule(frameinfo[0])
        if module and not module.__name__.startswith("requre"):
            return module


def get_class_that_defined_method(meth):
    """
    return class for given method meth

    :param meth: method where we would like to find class, where is methond defined
    :returns: class or None
    """
    # https://stackoverflow.com/questions/961048/get-class-that-defined-method
    for cls in inspect.getmro(meth.im_class):
        if meth.__name__ in cls.__dict__:
            return cls
    return None


def get_datafile_filename(obj, suffix=DEFAULT_SUFIX):
    """
    get default path for data files.
    It consist of 3 pieces: "location of test"/test_data/test_file_name/"test_id or function name"

    :param obj: object from which try to guess name of file and functioon
    :return: str with path where to store data_file
    """

    try:
        if isinstance(obj, Function):
            # pytest fixture
            current_fn_file_name = obj.module.__file__
        else:
            # try to get filename via class if possible (pytest way)
            current_fn_file_name = inspect.getfile(obj.__class__)
    except (AttributeError, TypeError):
        # try to get filename from object
        current_fn_file_name = inspect.getfile(obj)

    real_path_dir = Path(current_fn_file_name).parent.absolute()
    test_file_name = Path(current_fn_file_name).name.rsplit(".", 1)[0]
    old_test_name = None
    try:
        # try to use object.id() function (it is defined inside pytest unittests)
        old_test_name = test_name = obj.id()
        # use just class and function as indentifier of file
        cls_name = obj.__class__.__name__
        test_name = cls_name + test_name.rsplit(cls_name, 1)[1]
    except AttributeError:
        try:
            if isinstance(obj, Function):
                # pytest fixture
                test_name = obj.name
            else:
                # try to use __name__ of the object (typically name of function)
                test_name = obj.__name__
        except AttributeError:
            # if not possible, use this name as name of data file
            test_name = "static_test_data_name"
    testdata_dirname = real_path_dir / RELATIVE_TEST_DATA_DIRECTORY / test_file_name
    # BACKWARD COMPATIBILITY (read mode): check if a file using a full self.id is present
    if old_test_name:
        old_store_path = testdata_dirname / f"{old_test_name}.{suffix}"
        if old_store_path.exists():
            return old_store_path
    return testdata_dirname / f"{test_name}.{suffix}"



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import builtins
import os
import unittest
from shlex import split

from requre.import_system import UpgradeImportSystem, ReplaceType

SELECTOR = os.path.basename(__file__).rsplit(".", 1)[0]


class ContextManager(unittest.TestCase):
    def test_no_filter(self):
        old_imports = builtins.__import__
        with UpgradeImportSystem():
            import tempfile

            self.assertNotIn("decorated", tempfile.mktemp())
            self.assertIn("/tmp", tempfile.mktemp())
        assert builtins.__import__ == old_imports

    def test_filter_already_in_sys(self):
        old_imports = builtins.__import__
        with UpgradeImportSystem().upgrade(
            what="tempfile.mktemp",
            replace_type=ReplaceType.DECORATOR,
            replacement=lambda x: lambda: f"decorated {x()}",
        ):
            import tempfile

            self.assertIn("decorated", tempfile.mktemp())
            self.assertIn("/tmp", tempfile.mktemp())
        assert old_imports == builtins.__import__

    def test_filter_import_random(self):
        old_imports = builtins.__import__

        with UpgradeImportSystem().upgrade(
            what="random.random",
            replace_type=ReplaceType.DECORATOR,
            replacement=lambda x: lambda: f"decorated {x()}",
        ):
            import random

            self.assertIn("decorated", random.random())
            self.assertIn("0.", random.random())
        self.assertIsInstance(random.random(), float)
        assert old_imports == builtins.__import__

    def test_filter_import_from(self):
        with UpgradeImportSystem().upgrade(
            what="shlex.split",
            replace_type=ReplaceType.DECORATOR,
            replacement=lambda x: lambda y: f"mocked {x(y)}",
            add_revert_list=["tests.test_context_manager.split"],
        ):
            self.assertEqual("mocked ['a', 'b']", split("a b"))
        self.assertEqual(["a", "b"], split("a b"))

    def test_filter_import_from_reverted(self):
        with UpgradeImportSystem().upgrade(
            what="shlex.split",
            replace_type=ReplaceType.DECORATOR,
            replacement=lambda x: lambda y: f"mocked {x(y)}",
        ):
            self.assertEqual("mocked ['a', 'b']", split("a b"))
        self.assertEqual(["a", "b"], split("a b"))



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging

import pytest

from requre.online_replacing import recording_requests
from requre.utils import StorageMode, get_datafile_filename

logger = logging.getLogger(__name__)


@pytest.fixture
def record_requests_fixture(request):
    storage_file = get_datafile_filename(request.node)
    with recording_requests(storage_file=storage_file) as cassette:
        mode_description = {
            StorageMode.read: "replaying",
            StorageMode.write: "recording",
            StorageMode.append: "appending",
            StorageMode.default: "in default mode",
        }[cassette.mode]
        logger.debug(
            f"Start requre {mode_description} with storage file: {storage_file}"
        )
        yield cassette
        cassette.dump()
        logger.debug(f"End requre {mode_description} with storage file: {storage_file}")



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import os
import unittest
import requests
import socket

from requre.helpers import record_requests
from requre.cassette import Cassette, StorageMode

PAGE = "http://example.com"


@record_requests
class TestWriteCls(unittest.TestCase):
    def cassette_teardown(self, cassette: Cassette):
        self.cassette = cassette
        assert cassette.storage_file
        assert cassette.storage_object
        sf = self.cassette.storage_file
        if sf and os.path.exists(sf):
            os.remove(sf)

    def test(self):
        requests.get(PAGE)


class TestWrite(unittest.TestCase):
    def cassette_setup(self, cassette: Cassette):
        sf = self.cassette.storage_file
        self.cassette.storage_object = {}
        if sf and os.path.exists(sf):
            os.remove(sf)

    def cassette_teardown(self, cassette: Cassette):
        self.cassette = cassette
        assert cassette.storage_file
        assert cassette.storage_object

    @record_requests
    def test(self):
        self.test_executed = True
        requests.get(PAGE)

    def tearDown(self):
        assert self.test_executed


@record_requests
class TestRead(unittest.TestCase):
    def cassette_setup(self, cassette: Cassette):
        # disable network for read mode to prove it read from file
        if cassette.mode == StorageMode.read:
            setattr(socket, "socket", lambda x: x)

    def cassette_teardown(self, cassette: Cassette):
        self.cassette = cassette
        assert cassette.storage_file
        assert cassette.storage_object

    def test(self):
        requests.get(PAGE)


@record_requests
class TestFail(unittest.TestCase):
    def test(self):
        self.test_executed = True

    def tearDown(self):
        assert self.test_executed



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT


class SingletonMeta(type):
    """
    Metaclass for singleton, eg. to have right one object for persistent storage
    """

    _instance = None

    def __call__(self):
        if self._instance is None:
            self._instance = super().__call__()
        return self._instance



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging
import pickle
import warnings
from typing import Any, Dict, Optional

import yaml

from requre.objects import ObjectStorage
from requre.simple_object import Simple, Tuple, Void

logger = logging.getLogger(__name__)
GUESS_STR = "guess_type"


class Guess(Simple):
    """
    This class is able to store/read all simple types in requre
    Void, Simple, ObjectStorage (pickle)
    It select proper type if possible.

    Warning: when it uses ObjectStorage, it may lead some secrets to pickled objects
    and it could be hidden inside object representation.
    """

    @staticmethod
    def guess_type(value):
        if isinstance(value, tuple):
            return Tuple
        try:
            # Try to use type for storing simple output (list, dict, str, nums, etc...)
            yaml.safe_dump(value)
            return Simple
        except Exception:
            try:
                # Try to store anything serializable via pickle module
                pickle.dumps(value)
                return ObjectStorage
            except Exception:
                # do not store anything if not possible directly
                warnings.warn(
                    "Guess class - nonserializable return object - "
                    f"Using supressed output, are you sure? {value}"
                )
                return Void

    def write(self, obj: Any, metadata: Optional[Dict] = None) -> Any:
        """
        Write the object representation to storage
        Internally it will use self.to_serializable
        method to get serializable object representation

        :param obj: some object
        :param metadata: store metedata to object
        :return: same obj
        """
        object_serialization_type = self.guess_type(obj)
        metadata[GUESS_STR] = object_serialization_type.__name__
        instance = object_serialization_type(
            store_keys=self.store_keys,
            cassette=self.get_cassette(),
            storage_object_kwargs=self.storage_object_kwargs,
        )
        return instance.write(obj, metadata=metadata)

    def read(self):
        """
        Crete object representation of serialized data in persistent storage
        Internally it will use self.from_serializable method transform object

        :return: proper object
        """
        data = self.get_cassette()[self.store_keys]
        guess_type = self.get_cassette().data_miner.metadata[GUESS_STR]
        if guess_type == ObjectStorage.__name__:
            return pickle.loads(data)
        elif guess_type == Simple.__name__:
            return data
        elif guess_type == Void.__name__:
            return data
        elif guess_type == Tuple.__name__:
            return tuple(data)
        else:
            raise ValueError(
                f"Unsupported type of stored object inside cassette: {guess_type}"
            )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from requre.import_system import UpgradeImportSystem
from requre.simple_object import Simple

FILTERS = UpgradeImportSystem().decorate("time.sleep", Simple.decorator_plain())



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
This file is used by users to import special recording decorators or context managers.

* Can contain pre-made helpers for modules/functions with side effects.
* Can contain helpers consisting from multiple replacements
  so user can have only one for a group of related replacements that are always used together.
"""

from requre.helpers.git.helper import record_git_module
from requre.helpers.requests_response import record_requests, recording_requests
from requre.helpers.tempfile import record_tempfile_module

__all__ = [
    record_requests.__name__,
    recording_requests.__name__,
    record_tempfile_module.__name__,
    record_git_module.__name__,
]



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import os
import shutil

from requre import RequreTestCase
from requre.utils import get_datafile_filename


class CheckBaseTestClass(RequreTestCase):
    def tearDown(self):
        super().tearDown()
        data_file_path = get_datafile_filename(self)
        self.assertTrue(os.path.exists(data_file_path))
        # use just class and test name instead of full ID
        self.assertIn(".".join(self.id().split(".")[-2:]), data_file_path.name)
        self.assertIn("test_data", str(data_file_path))
        shutil.rmtree(os.path.dirname(get_datafile_filename(self)))

    def test(self):
        pass



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from typing import Optional

from requre.cassette import Cassette
from requre.helpers.files import StoreFiles
from requre.helpers.git.fetchinfo import FetchInfoStorageList
from requre.helpers.git.pushinfo import PushInfoStorageList
from requre.helpers.git.repo import Repo
from requre.record_and_replace import (
    make_generic,
    replace_module_match_with_multiple_decorators,
)


@make_generic
def record_git_module(
    _func=None,
    cassette: Optional[Cassette] = None,
):
    decorators = [
        (
            "git.repo.base.Repo.clone_from",
            StoreFiles.where_arg_references(
                key_position_params_dict={"to_path": 2},
                output_cls=Repo,
                cassette=cassette,
            ),
        ),
        (
            "git.remote.Remote.push",
            PushInfoStorageList.decorator_plain(),
        ),
        ("git.remote.Remote.fetch", FetchInfoStorageList.decorator_plain()),
        ("git.remote.Remote.pull", FetchInfoStorageList.decorator_plain()),
    ]
    record_git_decorator = replace_module_match_with_multiple_decorators(
        *decorators,
        cassette=cassette,
    )

    if _func is not None:
        return record_git_decorator(_func)
    else:
        return record_git_decorator



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from typing import Any, Dict, Optional

from requre.objects import ObjectStorage


class Simple(ObjectStorage):
    """
    Use this object, when your output is directly YAML serializable, basic objects e.g.
       string, number, boolean, list, dict of these basic objects
    """

    def from_serializable(self, data: Any) -> Any:
        return data

    def to_serializable(self, obj: Any) -> Any:
        return obj


class Void(Simple):
    """
    Use this object, when don't want to store output data of function.
    Could be used for debugging purposes, to see caller arguments
    or output is not important or unable to serialize anyhow.
    """

    def write(self, obj: Any, metadata: Optional[Dict] = None) -> Any:
        """
        Write the object representation to storage
        Internally it will use self.to_serializable()
        method to get serializable object representation

        :param obj: some object
        :param metadata: store metedata to object
        :return: same obj
        """
        self.get_cassette().store(
            self.store_keys,
            f">>>>> Requre output supressed by using {self.__class__.__name__}",
            metadata=metadata,
        )
        return obj


class Tuple(Simple):
    def to_serializable(self, obj):
        return list(obj)

    def from_serializable(self, data):
        return tuple(data)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import math
import unittest
from pathlib import Path

import requests

from requre.cassette import Cassette, StorageMode
from requre.simple_object import Simple
from requre.online_replacing import (
    record_requests_for_all_methods,
    apply_decorator_to_all_methods,
    replace_module_match,
)


@apply_decorator_to_all_methods(
    replace_module_match(what="math.sin", decorate=Simple.decorator_plain())
)
@record_requests_for_all_methods()
class DecoratorClassApplyMultipleDecorators(unittest.TestCase):
    def use_math_and_requests(self):
        sin_output = math.sin(1.5)
        self.assertAlmostEqual(0.9974949866040544, sin_output, delta=0.0005)

        response = requests.get("http://example.com")
        self.assertIn("This domain is for use", response.text)

    def test_read(self, cassette: Cassette):
        if cassette.mode == StorageMode.read:
            self.assertIn("math", cassette.storage_object)
            self.assertIn("requests.sessions", cassette.storage_object)
        # move the following call to the beginning when regenerating
        self.use_math_and_requests()

    def test_write(self, cassette: Cassette):
        self.assertIsNotNone(cassette)

        self.use_math_and_requests()

        self.assertIn("math", cassette.storage_object)
        self.assertIn("requests.sessions", cassette.storage_object)

        self.assertFalse(Path(cassette.storage_file).exists())
        cassette.dump()
        self.assertTrue(Path(cassette.storage_file).exists())
        Path(cassette.storage_file).unlink()


@apply_decorator_to_all_methods(
    replace_module_match(what="math.sin", decorate=Simple.decorator_plain())
)
@record_requests_for_all_methods()
class DecoratorClassApplyMultipleDecoratorsCassetteSetupTearDown(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.count_cassette_setup_triggered = 0
        self.count_cassette_teardown_triggered = 0

    def tearDown(self) -> None:
        super().tearDown()
        self.assertEqual(self.count_cassette_setup_triggered, 1)
        self.assertEqual(self.count_cassette_teardown_triggered, 1)

    def cassette_setup(self, cassette: Cassette):
        self.assertIsNotNone(cassette)
        self.count_cassette_setup_triggered += 1

    def cassette_teardown(self, cassette: Cassette):
        self.assertIsNotNone(cassette)
        self.count_cassette_teardown_triggered += 1

    def use_math_and_requests(self):
        sin_output = math.sin(1.5)
        self.assertAlmostEqual(0.9974949866040544, sin_output, delta=0.0005)

        response = requests.get("http://example.com")
        self.assertIn("This domain is for use", response.text)

    def test_read(self, cassette: Cassette):
        if cassette.mode == StorageMode.read:
            self.assertEqual(self.count_cassette_setup_triggered, 1)
            self.assertEqual(self.count_cassette_teardown_triggered, 0)
            self.assertIn("math", cassette.storage_object)
            self.assertIn("requests.sessions", cassette.storage_object)
        # move the following call to the beginning when regenerating
        self.use_math_and_requests()

    def test_write(self, cassette: Cassette):
        self.assertIsNotNone(cassette)
        self.assertEqual(self.count_cassette_setup_triggered, 1)
        self.assertEqual(self.count_cassette_teardown_triggered, 0)

        self.use_math_and_requests()

        self.assertIn("math", cassette.storage_object)
        self.assertIn("requests.sessions", cassette.storage_object)

        self.assertFalse(Path(cassette.storage_file).exists())
        cassette.dump()
        self.assertTrue(Path(cassette.storage_file).exists())
        Path(cassette.storage_file).unlink()


@record_requests_for_all_methods()
class DecoratorRecordRequestForAllMethodsCassetteSetupTearDown(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.count_cassette_setup_triggered = 0
        self.count_cassette_teardown_triggered = 0

    def tearDown(self) -> None:
        super().tearDown()
        self.assertEqual(self.count_cassette_setup_triggered, 1)
        self.assertEqual(self.count_cassette_teardown_triggered, 1)

    def cassette_setup(self, cassette: Cassette):
        self.assertIsNotNone(cassette)
        self.count_cassette_setup_triggered += 1

    def cassette_teardown(self, cassette: Cassette):
        self.assertIsNotNone(cassette)
        self.count_cassette_teardown_triggered += 1

    def use_requests(self):
        response = requests.get("http://example.com")
        self.assertIn("This domain is for use", response.text)

    def test_read(self, cassette: Cassette):
        if cassette.mode == StorageMode.read:
            self.assertEqual(self.count_cassette_setup_triggered, 1)
            self.assertEqual(self.count_cassette_teardown_triggered, 0)
            self.assertIn("requests.sessions", cassette.storage_object)
        # move the following call to the beginning when regenerating
        self.use_requests()

    def test_write(self, cassette: Cassette):
        self.assertIsNotNone(cassette)
        self.assertEqual(self.count_cassette_setup_triggered, 1)
        self.assertEqual(self.count_cassette_teardown_triggered, 0)

        self.use_requests()

        self.assertIn("requests.sessions", cassette.storage_object)

        self.assertFalse(Path(cassette.storage_file).exists())
        cassette.dump()
        self.assertTrue(Path(cassette.storage_file).exists())
        Path(cassette.storage_file).unlink()


@record_requests_for_all_methods()
class DecoratorRecordRequestForAllMethods(unittest.TestCase):
    def use_requests(self):
        response = requests.get("http://example.com")
        self.assertIn("This domain is for use", response.text)

    def test_read(self, cassette: Cassette):
        if cassette.mode == StorageMode.read:
            self.assertIn("requests.sessions", cassette.storage_object)
        self.use_requests()

    def test_write(self, cassette: Cassette):
        self.assertIsNotNone(cassette)

        self.use_requests()

        self.assertIn("requests.sessions", cassette.storage_object)

        self.assertFalse(Path(cassette.storage_file).exists())
        cassette.dump()
        self.assertTrue(Path(cassette.storage_file).exists())
        Path(cassette.storage_file).unlink()


@record_requests_for_all_methods
class DecoratorRecordRequestForAllMethodsWithoutParenthesis(unittest.TestCase):
    def use_requests(self):
        response = requests.get("http://example.com")
        self.assertIn("This domain is for use", response.text)

    def test_read(self, cassette: Cassette):
        if cassette.mode == StorageMode.read:
            self.assertIn("requests.sessions", cassette.storage_object)
        self.use_requests()

    def test_write(self, cassette: Cassette):
        self.assertIsNotNone(cassette)

        self.use_requests()

        self.assertIn("requests.sessions", cassette.storage_object)

        self.assertFalse(Path(cassette.storage_file).exists())
        cassette.dump()
        self.assertTrue(Path(cassette.storage_file).exists())
        Path(cassette.storage_file).unlink()


@record_requests_for_all_methods(regexp_method_pattern="test_regex.*")
class DecoratorRecordRequestForAllMethodsDifferentRegex(unittest.TestCase):
    def some_method(self):
        pass

    def use_requests(self):
        response = requests.get("http://example.com")
        self.assertIn("This domain is for use", response.text)

    def test_without_decorator(self, cassette: Cassette = None):
        self.assertIsNone(cassette)

    def test_regex_read(self, cassette: Cassette):
        if cassette.mode == StorageMode.read:
            self.assertIn("requests.sessions", cassette.storage_object)
        self.use_requests()

    def test_regex_write(self, cassette: Cassette):
        self.assertIsNotNone(cassette)

        self.use_requests()

        self.assertIn("requests.sessions", cassette.storage_object)

        self.assertFalse(Path(cassette.storage_file).exists())
        cassette.dump()
        self.assertTrue(Path(cassette.storage_file).exists())
        Path(cassette.storage_file).unlink()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import random


def hello(text="Hi! "):
    return text


def inc(value):
    return value + 1


class dynamic:
    def __init__(self):
        self.text = "called"

    def __getattr__(self, item):
        return lambda: item.upper()


def random_number():
    return random.random()


def another_random_number():
    return random.random()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import unittest
import math
import os
from requre.objects import ObjectStorage
from requre.utils import StorageMode
from requre.guess_object import Guess, GUESS_STR
from requre.simple_object import Void, Simple
from requre.online_replacing import apply_decorator_to_all_methods, replace_module_match
from requre.cassette import Cassette
from tests.testbase import BaseClass
from tests.test_object import OwnClass
import sys


class Unit(BaseClass):
    def testVoid(self):
        self.assertEqual(Void, Guess.guess_type(sys.__stdout__))

    def testSimple(self):
        self.assertEqual(Simple, Guess.guess_type("abc"))
        self.assertEqual(Simple, Guess.guess_type({1: 2, "a": "b"}))
        self.assertNotEqual(ObjectStorage, Guess.guess_type({1: 2, "a": "b"}))

    def testPickle(self):
        self.assertEqual(ObjectStorage, Guess.guess_type(OwnClass(1)))
        self.assertNotEqual(Void, Guess.guess_type(OwnClass(1)))
        self.assertNotEqual(Simple, Guess.guess_type(OwnClass(1)))


def obj_return(num, obj):
    _ = num
    return obj


class Store(BaseClass):
    def testFunctionDecorator(self):
        """
        Check if it is able to guess and store/restore proper values with types
        """
        decorated_own = Guess.decorator(cassette=self.cassette, item_list=[0])(
            obj_return
        )
        before1 = decorated_own(1, "abc")
        before2 = decorated_own(2, OwnClass(2))
        before3 = decorated_own(2, OwnClass(3))
        before4 = decorated_own(3, sys.__stdin__)
        before5 = decorated_own(
            4,
            (
                1,
                2,
            ),
        )
        self.cassette.dump()
        self.cassette.mode = StorageMode.read

        after2 = decorated_own(2, OwnClass(2))
        self.assertEqual(
            self.cassette.data_miner.metadata[GUESS_STR], ObjectStorage.__name__
        )
        after3 = decorated_own(2, OwnClass(3))
        self.assertEqual(
            self.cassette.data_miner.metadata[GUESS_STR], ObjectStorage.__name__
        )
        after1 = decorated_own(1, "abc")
        self.assertEqual(self.cassette.data_miner.metadata[GUESS_STR], Simple.__name__)
        after4 = decorated_own(3, sys.__stdin__)
        self.assertEqual(self.cassette.data_miner.metadata[GUESS_STR], Void.__name__)
        after5 = decorated_own(4, (3,))
        self.assertEqual(
            after5,
            (
                1,
                2,
            ),
        )
        self.assertTrue(isinstance(after5, tuple))

        self.assertEqual(before1, after1)
        self.assertEqual(before2.__class__.__name__, after2.__class__.__name__)
        self.assertEqual(before3.__class__.__name__, after3.__class__.__name__)
        self.assertEqual(after2.__class__.__name__, "OwnClass")
        self.assertEqual(before4.__class__.__name__, "TextIOWrapper")
        self.assertEqual(after4.__class__.__name__, "str")
        self.assertEqual(before5.__class__.__name__, "tuple")
        self.assertEqual(before5, after5)


@apply_decorator_to_all_methods(replace_module_match(what="math.sin"))
class ApplyDefaultDecorator(unittest.TestCase):
    SIN_OUTPUT = 0.9974949866040544

    def cassette_setup(self, cassette):
        self.assertEqual(cassette.storage_object, {})

    def cassette_teardown(self, cassette):
        os.remove(cassette.storage_file)

    def test(self, cassette: Cassette):
        math.sin(1.5)
        self.assertEqual(len(cassette.storage_object["math"]["sin"]), 1)
        self.assertAlmostEqual(
            self.SIN_OUTPUT,
            cassette.storage_object["math"]["sin"][0]["output"],
            delta=0.0005,
        )
        self.assertEqual(
            cassette.storage_object["math"]["sin"][0]["metadata"][GUESS_STR],
            "Simple",
        )
